import numpy as np
import math
from numpy import cos, sin
from scipy.interpolate import RectBivariateSpline
from scipy.ndimage import rotate

import proper
import falco
from . import check
from falco.util import ceil_even, cosd, sind


def _init_proper(Dmask, dx, centering):
    """Initialize PROPER for generating mask representations."""
    check.centering(centering)

    # number of points across output array:
    if centering == "pixel":
        # Sometimes requires two more pixels when pixel centered.
        # Same size as width when interpixel centered.
        Narray = ceil_even(Dmask / dx + 0.5)
    else:
        Narray = ceil_even(Dmask / dx + 0.0)  # Same size as width when interpixel centered.

    wl_dummy = 1e-6  # wavelength (m); Dummy value--no propagation here, so not used.

    return proper.prop_begin(Narray * dx, wl_dummy, Narray, 1.0)


def falco_gen_DM_stop(dx, diamMask, centering):
    """
    Make a circular aperture to place centered on the beam at a DM.

    Parameters
    ----------
    dx : float
        spatial resolution for a pixel. Any units as long as they match that of
        diamMask
    diamMask : float
        diameter of the aperture mask. Any units as long as they match that of
        dx
    centering :
        centering of beam in array. Either 'pixel' or 'interpixel'

    Returns
    -------
    numpy ndarray
        2-D square array of a circular stop at a DM. Cropped down to the 
        smallest even-sized array with no extra zero padding.

    """
    check.real_positive_scalar(dx, 'dx', TypeError)
    check.real_positive_scalar(diamMask, 'Dmask', TypeError)
    check.centering(centering)

    wf = _init_proper(diamMask, dx, centering)

    # 0 shift for pixel-centered pupil, or -dx shift for interpixel centering
    cshift = -dx / 2 * (centering == "interpixel")

    # Outer diameter of aperture
    proper.prop_circular_aperture(wf, diamMask/2., cshift, cshift)

    return np.fft.ifftshift(np.abs(wf.wfarr))


def falco_gen_pupil_Roman_CGI_20200513(Nbeam, centering, changes={}):
    """
    Generate WFIRST pupil CGI-20191009.

    Function to generate WFIRST pupil CGI-20191009. Options to change
    the x- or y-shear, clocking, or magnification via keys in the dict
    called changes.

    Parameters
    ----------
    Nbeam : float, int
        Number of pixels across the diameter of the aperture.
    centering : string
        String specifying the centering of the output array
    changes : dict
        Optional dictionary of specifications for altering the nominal pupil

    Returns
    -------
    pupil : numpy ndarray
        2-D amplitude map of the WFIRST pupil CGI-20191009
    """
    check.real_positive_scalar(Nbeam, 'Nbeam', TypeError)
    check.centering(centering)
    check.is_dict(changes, 'changes')

    # Define the best-fit values for ellipses and rectangles (DO NOT CHANGE)

    primaryRadiusYpixels = 4027.25
    ODpixels = 2*primaryRadiusYpixels

    primaryRadiusX = 3990.0/ODpixels
    primaryRadiusY = primaryRadiusYpixels/ODpixels
    primaryCenterX = 0.0
    primaryCenterY = 0.0

    secondaryRadiusX = 1209.65/ODpixels
    secondaryRadiusY = 1220.0/ODpixels
    secondaryCenterX = 0.0/ODpixels
    secondaryCenterY = -2.95/ODpixels

    strutEndVecX1 = np.array([843.9, 728.0, 47.5, -192.0, -676.0, -816.65])/ODpixels
    strutEndVecY1 = np.array([550.85, 580.35, -970.65, -1097.15, 605.55, 458.85])/ODpixels

    strutEndVecX2 = np.array([1579.9, 3988.0, 2429.7, -2484.0, -3988.0, -1572.65])/ODpixels
    strutEndVecY2 = np.array([3866.85, -511.65, -3214.65, -3256.15, -504.45, 3866.85])/ODpixels

    strutCenterVecX = (strutEndVecX1 + strutEndVecX2)/2.0
    strutCenterVecY = (strutEndVecY1 + strutEndVecY2)/2.0

    strutWidthVec = np.array([257.0, 259.0, 258.0, 258.0, 259.0, 257.0])/ODpixels

    strutAngleVec = np.arctan2(strutEndVecY2-strutEndVecY1, strutEndVecX2-strutEndVecX1)*(180/np.pi)

    tabRadiusVecX = np.array([1343.0, 1343.0, 1364.0])/ODpixels
    tabRadiusVecY = np.array([1353.0, 1353.0, 1374.0])/ODpixels
    tabCenterVecX = np.array([0.0, 0.0, 0.0])/ODpixels
    tabCenterVecY = np.array([53.85, 53.85, 67.6])/ODpixels

    lStrut = 0.55

    deltaAngle = 2.5*np.pi/16
    angTabStart = np.array([0.616 - deltaAngle/2.0, 2.54 - deltaAngle/2.0, -1.57 - deltaAngle/2.0])
    angTabEnd = np.array([0.616 + deltaAngle/2.0, 2.54 + deltaAngle/2.0, -1.57 + deltaAngle/2.0])

    xcStrutVec = strutCenterVecX
    ycStrutVec = strutCenterVecY
    angStrutVec = strutAngleVec
    wStrutVec = strutWidthVec

    ## Changes to the pupil

    ### (Optional) Lyot stop mode (concentric, circular ID and OD)
    if not 'flagLyot' in changes: 
        flagLyot = False
    else:
        flagLyot = changes['flagLyot']
    if flagLyot:
        if 'ID' in changes:
            ID = changes['ID']
        else:
            raise ValueError('ID must be defined for Lyot stop generation mode.')

        if 'OD' in changes:
            OD = changes['OD']
        else:
            raise ValueError('OD must be defined for Lyot stop generation mode.')

    # Oversized strut features: overwrite defaults if values specified
#    if 'OD' in changes: OD = changes["OD"]
#    if 'ID' in changes: ID = changes["ID"]
    if 'wStrut' in changes: wStrutVec = np.ones(6)*changes["wStrut"]
    if 'wStrutVec' in changes: wStrutVec = changes["wStrutVec"]

    # Padding values for obscuration
    # Defaults of Bulk Changes: (All length units are pupil diameters. All angles are in degrees.)
    if not 'xShear' in changes: changes["xShear"] = 0
    if not 'yShear' in changes: changes["yShear"] = 0
    if not 'magFac' in changes: changes["magFac"] = 1.0
    if not 'clock_deg' in changes: changes["clock_deg"] = 0.0
    if not 'flagRot180' in changes: changes["flagRot180"] = False

    # Defaults for obscuration padding: (All length units are pupil diameters.)
    if not 'pad_all' in changes: changes["pad_all"] = 0.
    if not 'pad_strut' in changes: changes["pad_strut"] = 0.
    if not 'pad_COBS' in changes: changes["pad_COBS"] = 0.
    if not 'pad_COBStabs' in changes: changes["pad_COBStabs"] = 0.
    if not 'pad_OD' in changes: changes["pad_OD"] = 0.

    # Values to use for bulk clocking, magnification, and translation
    xShear = changes["xShear"]  # - xcOD;
    yShear = changes["yShear"]  # - ycOD;
    magFac = changes["magFac"]
    clock_deg = changes["clock_deg"]
    clockRad = clock_deg*(np.pi/180.)
    flagRot180 = changes["flagRot180"]

    # Padding values. (pad_all is added to all the rest)
    pad_all = changes["pad_all"]     #0.2/100; # Uniform padding on all features
    pad_strut = changes["pad_strut"] + pad_all
    pad_COBS = changes["pad_COBS"] + pad_all
    pad_COBStabs = changes["pad_COBStabs"] + pad_all
    pad_OD = changes["pad_OD"] + pad_all  # Radial padding at the edge

    # Rotation matrix used on center coordinates.
    rotMat = np.array([[math.cos(clockRad), -math.sin(clockRad)],
                       [math.sin(clockRad), math.cos(clockRad)]])

    # Nominal Mask Coordinates
    if centering.lower() == 'pixel':
        Narray = ceil_even(Nbeam*np.max(2*np.abs((xShear, yShear))) + Nbeam*np.max([magFac, 1.0]) + 1)
    elif centering.lower() == 'interpixel':
        Narray = ceil_even(Nbeam*np.max(2*np.abs((xShear, yShear))) + Nbeam*np.max([magFac, 1.0]))

    if centering.lower() == 'interpixel':
        xs = np.linspace(-(Narray-1)/2, (Narray-1)/2, Narray)/Nbeam
    else:
        xs = np.linspace(-(Narray/2), Narray/2-1, Narray)/Nbeam

    [XS, YS] = np.meshgrid(xs, xs)

    # Proper Setup Values
    Dbeam = 1  # Diameter of aperture, normalized to itself
    wl = 1e-6  # wavelength (m); Dummy value--no propagation here, so not used.
    bdf = Nbeam/Narray  # beam diameter factor in output array
    dx = Dbeam/Nbeam
    nSubsamples = 101

    if centering.lower() in ('interpixel', 'even'):
        cshift = -dx/2.0
    elif centering.lower() in ('pixel', 'odd'):
        cshift = 0
        if flagRot180:
            cshift = -dx

    # INITIALIZE PROPER
    bm = proper.prop_begin(Dbeam, wl, Narray, bdf)
    proper.prop_set_antialiasing(nSubsamples)
    
    # Struts
    for iStrut in range(6):
        angDeg = angStrutVec[iStrut] + clock_deg  # degrees
        wStrut = magFac*(wStrutVec[iStrut] + 2*pad_strut)
        lStrutIn = magFac*lStrut
        xc = magFac*(xcStrutVec[iStrut])
        yc = magFac*(ycStrutVec[iStrut])
        cxy = rotMat @ np.array([xc, yc]).reshape((2, 1))
        xc = cxy[0] + xShear
        yc = cxy[1] + yShear
        proper.prop_rectangular_obscuration(bm, lStrutIn, wStrut, xc+cshift,
                                            yc+cshift, ROTATION=angDeg)

    if not flagLyot:

        # PRIMARY MIRROR (OUTER DIAMETER)
        ra_OD_x = magFac*(primaryRadiusX-pad_OD)
        ra_OD_y = magFac*(primaryRadiusY-pad_OD)
        cx_OD = magFac*primaryCenterX
        cy_OD = magFac*primaryCenterY
        cxy = rotMat @ np.array([cx_OD, cy_OD]).reshape((2, 1))
        cx_OD = cxy[0] + xShear + cshift
        cy_OD = cxy[1] + yShear + cshift
        proper.prop_elliptical_aperture(bm, ra_OD_x, ra_OD_y, cx_OD, cy_OD,
                                        ROTATION=-clock_deg)

        # inputs = {}
        # inputs["Narray"] = Narray
        # inputs["Nbeam"] = Nbeam
        # inputs["radiusX"] = ra_OD_x
        # inputs["radiusY"] = ra_OD_y
        # inputs["clockingDegrees"] = clock_deg
        # inputs["centering"] = centering
        # inputs["xShear"] = cx_OD
        # inputs["yShear"] = cy_OD
        # primaryAperture = falco_gen_ellipse(inputs)

        # SECONDARY MIRROR (INNER DIAMETER)
        ra_ID_x = magFac*(secondaryRadiusX + pad_COBS)
        ra_ID_y = magFac*(secondaryRadiusY + pad_COBS)
        cx_ID = magFac*secondaryCenterX
        cy_ID = magFac*secondaryCenterY
        cxy = rotMat @ np.array([cx_ID, cy_ID]).reshape((2, 1))
        cx_ID = cxy[0] + xShear + cshift
        cy_ID = cxy[1] + yShear + cshift
        proper.prop_elliptical_obscuration(bm, ra_ID_x, ra_ID_y, cx_ID, cy_ID,
                                           ROTATION=-clock_deg)

        # inputs = {}
        # inputs["Narray"] = Narray
        # inputs["Nbeam"] = Nbeam
        # inputs["radiusX"] = ra_ID_x
        # inputs["radiusY"] = ra_ID_y
        # inputs["clockingDegrees"] = clock_deg
        # inputs["centering"] = centering
        # inputs["xShear"] = cx_ID
        # inputs["yShear"] = cy_ID
        # secondaryObscuration = 1 - falco_gen_ellipse(inputs)

        # Tabs where Struts Meet Secondary Mirror
        nTabs = 3
        tabCube = np.ones((Narray, Narray, nTabs))

        for iTab in range(nTabs):
            cobsTabsMask = np.zeros((Narray, Narray))

            xyShear = rotMat @ np.array([tabCenterVecX[iTab],
                                         tabCenterVecY[iTab]]).reshape((2, 1))
            
            XSnew = (XS + magFac*xyShear[0]) - xShear
            YSnew = (YS + magFac*xyShear[1]) - yShear
            THETAS = np.arctan2(YSnew, XSnew)

            ang1 = angTabStart[iTab] + clockRad
            ang2 = angTabEnd[iTab] + clockRad
            while (ang2 > 2*np.pi) and (ang1 > 2*np.pi):
                ang1 = ang1 - 2*np.pi
                ang2 = ang2 - 2*np.pi
            while (ang2 < -2*np.pi) and (ang1 < -2*np.pi):
                ang1 = ang1 + 2*np.pi
                ang2 = ang2 + 2*np.pi
            if ((ang2 < 2*np.pi) and (ang2 > np.pi)) and ((ang1 < np.pi) or ang1 > 0):
                THETAS[THETAS < 0] = THETAS[THETAS < 0] + 2*np.pi
            elif (ang2 > 2*np.pi) and ((ang1 > np.pi) and (ang1 < 2*np.pi)):
                THETAS = THETAS + 2*np.pi
            elif (ang1 < -np.pi) and (ang2 > -np.pi):
                THETAS[THETAS > 0] = THETAS[THETAS > 0] - 2*np.pi
            elif (ang1 < -np.pi) and (ang2 < -np.pi):
                THETAS = THETAS - 2*np.pi

            cobsTabsMask[(THETAS <= ang2) & (THETAS >= ang1)] = 1.0
        
            # if(angTabStart[iTab] > angTabEnd[iTab]):
            #     cobsTabsMask[(THETAS >= (angTabEnd[iTab]+clockRad)) &
            #                  (THETAS <= (angTabStart[iTab]+clockRad))] = 1.0
            # else:
            #     cobsTabsMask[(THETAS <= (angTabEnd[iTab]+clockRad)) &
            #                  (THETAS >= (angTabStart[iTab]+clockRad))] = 1.0

            # Full ellipse to be multiplied by the mask to get just tabs
            cx_tab = magFac*tabCenterVecX[iTab]
            cy_tab = magFac*tabCenterVecY[iTab]
            cxy = rotMat @ np.array([cx_tab, cy_tab]).reshape((2, 1))
            cx_tab = cxy[0] + xShear
            cy_tab = cxy[1] + yShear
            tabRadiusX = magFac*(tabRadiusVecX[iTab] + pad_COBStabs)
            tabRadiusY = magFac*(tabRadiusVecY[iTab] + pad_COBStabs)
            bm2 = proper.prop_begin(Dbeam, wl, Narray, bdf)
            proper.prop_set_antialiasing(nSubsamples)
            proper.prop_elliptical_obscuration(bm2, tabRadiusX, tabRadiusY,
                                               cx_tab+cshift, cy_tab+cshift,
                                               ROTATION=-clock_deg)
            tabEllipse = 1 - np.fft.ifftshift(np.abs(bm2.wfarr))

            # inputs = {}
            # inputs["Narray"] = Narray
            # inputs["Nbeam"] = Nbeam
            # inputs["radiusX"] = tabRadiusX
            # inputs["radiusY"] = tabRadiusY
            # inputs["clockingDegrees"] = clock_deg
            # inputs["centering"] = centering
            # inputs["xShear"] = cx_tab
            # inputs["yShear"] = cy_tab
            # tabEllipse = falco_gen_ellipse(inputs)

            # clear inputs
            # inputs.Narray = Narray;
            # inputs.Nbeam = Nbeam #;% = 1219.0/ODpixels %Nbeam;
            # inputs.radiusX = tabRadiusX;
            # inputs.radiusY = tabRadiusY;
            # inputs.clockingDegrees = clock_deg;
            # inputs.centering = centering;
            # inputs.xShear = cx_tab;
            # inputs.yShear = cy_tab;
            # % inputs.magFac = magFac;
            # tabEllipse = falco_gen_ellipse(inputs);

            tabSector = cobsTabsMask*tabEllipse

            tabCube[:, :, iTab] = 1 - tabSector

        pupil = tabCube[:, :, 0] * tabCube[:, :, 1] * tabCube[:, :, 2] *\
            np.fft.ifftshift(np.abs(bm.wfarr))

    else:  # (Lyot stop mode)

        # OUTER DIAMETER
        ra_OD = magFac*(OD/2.)
        cx_OD = xShear
        cy_OD = yShear
        proper.prop_circular_aperture(bm, ra_OD, cx_OD+cshift, cy_OD+cshift)

        # INNER DIAMETER
        ra_ID = magFac*(ID/2.)
        cx_ID = xShear
        cy_ID = yShear
        proper.prop_circular_obscuration(bm, ra_ID, cx_ID+cshift, cy_ID+cshift)

        pupil = np.fft.ifftshift(np.abs(bm.wfarr))

    if(flagRot180):
        pupil = np.rot90(pupil, 2)

    return pupil


def falco_gen_pupil_WFIRST_CGI_180718(Nbeam, centering, changes={}):
    """
    Generate WFIRST pupil CGI-180718.

    Generate WFIRST pupil CGI-180718. Options to change the x- or y-shear,
    clocking, or magnification via the keyword argument changes.

    Parameters
    ----------
    mp: falco.config.ModelParameters
        Structure of model parameters
    centering : string
        String specifying the centering of the output array
    Returns
    -------
    pupil : numpy ndarray
        2-D amplitude map of the WFIRST pupil CGI-180718
    """
    check.real_positive_scalar(Nbeam, 'Nbeam', TypeError)
    check.centering(centering)
    check.is_dict(changes, 'changes')

    OD = 1.000130208333333
    xcOD = 8.680555555555557e-06
    ycOD = 8.680555555555557e-06
    ID = 3.030133333333332e-01
    xcCOBS = -1.155555555555556e-04
    ycCOBS = -6.133333333333334e-04
    IDtabs = 3.144078947368421e-01
    xcCOBStabs = -1.973684210526340e-04
    ycCOBStabs = -6.250000000000000e-03

    wStrutVec = np.array([
     3.219259259259259e-02,
     3.219259259259259e-02,
     3.219259259259259e-02,
     3.219259259259258e-02,
     3.219259259259259e-02,
     3.219259259259259e-02,
    ])
    lStrut = 5.500000000000000e-01
    angStrutVec = np.array([
     4.308638741879215e+01,
     1.828091850580443e+01,
     -7.736372240624411e+01,
     7.746228722667239e+01,
     -1.833049685311381e+01,
     -4.310697246349373e+01,
    ])
    xcStrutVec = np.array([
     1.637164789600492e-01,
     3.311169704392094e-01,
     1.542050924925356e-01,
     -1.556442459316893e-01,
     -3.075636241385107e-01,
     -1.712399202747162e-01,
    ])
    ycStrutVec = np.array([
     2.695837795868052e-01,
     7.744558909460633e-03,
     -2.885875977555251e-01,
     -2.874651682155463e-01,
     -7.319997758726773e-04,
     2.748434070552074e-01])
    angTabStart = np.array([
     1.815774989921760e+00,
     -3.487710035839058e-01,
     -2.416523875732038e+00])
    angTabEnd = np.array([
     1.344727938801013e+00,
     -7.527300509955320e-01,
     -2.822938064533701e+00])

    # Oversized strut features: overwrite defaults if values specified
    if 'OD' in changes: OD = changes["OD"]
    if 'ID' in changes: ID = changes["ID"]
    if 'wStrut' in changes: wStrutVec = np.ones(6)*changes["wStrut"]
    if 'wStrutVec' in changes: wStrutVec = changes["wStrutVec"]

    # Padding values for obscuration
    # Defaults of Bulk Changes: (All length units are pupil diameters. 
    # All angles are in degrees.)
    if not 'xShear' in changes: changes["xShear"] = 0
    if not 'yShear' in changes: changes["yShear"] = 0
    if not 'magFac' in changes: changes["magFac"] = 1.0
    if not 'clock_deg' in changes: changes["clock_deg"] = 0.0
    if not 'flagRot180' in changes: changes["flagRot180"] = False

    # Defaults for obscuration padding: (All length units are pupil diameters.)
    if not 'pad_all' in changes: changes["pad_all"] = 0.
    if not 'pad_strut' in changes: changes["pad_strut"] = 0.
    if not 'pad_COBS' in changes: changes["pad_COBS"] = 0.
    if not 'pad_COBStabs' in changes: changes["pad_COBStabs"] = 0.
    if not 'pad_OD' in changes: changes["pad_OD"] = 0.

    # Values to use for bulk clocking, magnification, and translation
    xShear = changes["xShear"]  # - xcOD;
    yShear = changes["yShear"]  # - ycOD;
    magFac = changes["magFac"]
    clock_deg = changes["clock_deg"]
    clockRad = clock_deg*(np.pi/180.)
    flagRot180 = changes["flagRot180"]

    # Padding values. (pad_all is added to all the rest)
    pad_all = changes["pad_all"]     #0.2/100; # Uniform padding on all features
    pad_strut = changes["pad_strut"] + pad_all
    pad_COBS = changes["pad_COBS"] + pad_all
    pad_COBStabs = changes["pad_COBStabs"] + pad_all
    pad_OD = changes["pad_OD"] + pad_all  # Radial padding at the edge

    # Rotation matrix used on center coordinates.
    rotMat = np.array([[math.cos(clockRad), -math.sin(clockRad)],
                       [math.sin(clockRad), math.cos(clockRad)]])

    # Coordinates
    if centering.lower() in ('pixel', 'odd'):
        Narray = ceil_even(Nbeam*np.max(2*np.abs((xShear, yShear))) + magFac*(Nbeam+1))
    else:
        Narray = ceil_even(Nbeam*np.max(2*np.abs((xShear, yShear))) + magFac*Nbeam)

    if centering.lower() == 'interpixel':
        xs = np.linspace(-(Nbeam-1)/2, (Nbeam-1)/2, Nbeam)/Nbeam
    else:
        xs = np.linspace(-(Narray/2), Narray/2-1, Narray)/Nbeam

    [XS, YS] = np.meshgrid(xs, xs)

    # Proper Setup Values
    Dbeam = 1  # Diameter of aperture, normalized to itself
    wl   = 1e-6  # wavelength (m); Dummy value--no propagation here, so not used.
    bdf = Nbeam/Narray  # beam diameter factor in output array
    dx = Dbeam/Nbeam

    if centering.lower() in ('interpixel', 'even'):
        cshift = -dx/2
    elif centering.lower() in ('pixel', 'odd'):
        cshift = 0
        if flagRot180:
            cshift = -dx

    # INITIALIZE PROPER
    bm = proper.prop_begin(Dbeam, wl, Narray, bdf)

    # PRIMARY MIRROR (OUTER DIAMETER)
    ra_OD = magFac*(OD/2 - pad_OD)
    cx_OD = magFac*xcOD
    cy_OD = magFac*ycOD
    cxy = np.matmul(rotMat, np.array([[cx_OD], [cy_OD]]))
    cx_OD = cxy[0]+xShear
    cy_OD = cxy[1]+yShear
    proper.prop_circular_aperture(bm, ra_OD, cx_OD+cshift, cy_OD+cshift)
    # bm = proper.prop_circular_aperture(bm, ra_OD, cx_OD+cshift, cy_OD+cshift)

    # SECONDARY MIRROR (INNER DIAMETER)
    ra_ID = magFac*(ID/2 + pad_COBS)
    cx_ID = magFac*xcCOBS
    cy_ID = magFac*ycCOBS
    cxy = np.matmul(rotMat, np.array([[cx_ID], [cy_ID]]))
    cx_ID = cxy[0]+xShear
    cy_ID = cxy[1]+yShear
    proper.prop_circular_obscuration(bm, ra_ID, cx_ID+cshift, cy_ID+cshift)
    # bm = proper.prop_circular_obscuration(bm, ra_ID, cx_ID+cshift, cy_ID+cshift)

    # Struts
    for istrut in range(6):
        angDeg = angStrutVec[istrut] + clock_deg  # degrees
        wStrut = magFac*(wStrutVec[istrut] + 2*pad_strut)
        lStrutIn = magFac*lStrut
        xc = magFac*(xcStrutVec[istrut])
        yc = magFac*(ycStrutVec[istrut])
        cxy = np.matmul(rotMat, np.array([[xc], [yc]]))
        xc = cxy[0]+xShear
        yc = cxy[1]+yShear
        proper.prop_rectangular_obscuration(bm, lStrutIn, wStrut, xc+cshift,\
                                            yc+cshift, ROTATION=angDeg)

    wf1 = bm.wfarr  # assign for later use

    # TABS ON SECONDARY MIRROR
    # Compute as new shape, and then multiply the obscuration with the rest of
    # the pupil.

    # SOFTWARE MASK:
    XSnew = (1/1*XS+xcCOBStabs)-xShear
    YSnew = (1/1*YS+ycCOBStabs)-yShear

    overSizeFac = 1.3
    cobsTabsMask = np.zeros([Narray, Narray])
    THETAS = np.arctan2(YSnew, XSnew)
    clockRad = np.deg2rad(clock_deg)

    if angTabStart[0] > angTabEnd[0]:
        msk1 = (XSnew**2 + YSnew**2) <= (overSizeFac*magFac*IDtabs/2)**2
        msk2 = np.logical_and(THETAS >= angTabEnd[0]+clockRad, THETAS <= angTabStart[0]+clockRad)
        msk3 = np.logical_and(THETAS >= angTabEnd[1]+clockRad, THETAS <= angTabStart[1]+clockRad)
        msk4 = np.logical_and(THETAS >= angTabEnd[2]+clockRad, THETAS <= angTabStart[2]+clockRad)
        cobsTabsMask[np.logical_and(msk1,np.logical_or(msk2,np.logical_or(msk3, msk4)))] = 1
    else:
        msk1 = (XSnew**2 + YSnew**2) <= (overSizeFac*magFac*IDtabs/2)**2
        msk2 = np.logical_and(THETAS <= angTabEnd[0]+clockRad, THETAS >= angTabStart[0]+clockRad)
        msk3 = np.logical_and(THETAS <= angTabEnd[1]+clockRad, THETAS >= angTabStart[1]+clockRad)
        msk4 = np.logical_and(THETAS <= angTabEnd[2]+clockRad, THETAS >= angTabStart[2]+clockRad)
        cobsTabsMask[np.logical_and(msk1, np.logical_or(msk2, np.logical_or(msk3, msk4)))] = 1


    # CIRCLE:
    # Initialize PROPER
    bm = proper.prop_begin(Dbeam, wl, Narray,bdf)

    # Full circle of COBS tabs--to be multiplied by the mask to get just tabs
    ra_tabs = magFac*(IDtabs/2 + pad_COBStabs)
    cx_tabs = magFac*(xcCOBStabs)
    cy_tabs = magFac*(ycCOBStabs)
    cxy = np.matmul(rotMat,np.array([[cx_tabs], [cy_tabs]]))
    cx_tabs = cxy[0]+xShear
    cy_tabs = cxy[1]+yShear

    # bm2 = prop_circular_obscuration(bm, ra_tabs,'XC',cx_tabs+cshift,'YC',cy_tabs+cshift)
    proper.prop_circular_obscuration(bm, ra_tabs,cx_tabs+cshift,cy_tabs+cshift)

    temp = 1-np.fft.ifftshift(np.abs(bm.wfarr))
    temp = cobsTabsMask*temp

    cobsTabs = 1-temp

    # Output
    pupil = cobsTabs*np.fft.ifftshift(np.abs(wf1))
    if(flagRot180):
        pupil = np.rot90(pupil, 2)

    return pupil


def falco_gen_SW_mask(inputs):
    """
    Function to generate binary (0-1) software masks for the focal plane. 
    This can be used as a field stop, or for making the scoring and correction 
    regions in the focal plane.

    Detailed description here

    Parameters
    ----------
    inputs: structure with several fields:
   -pixresFP: pixels per lambda_c/D
   -rhoInner: radius of inner FPM amplitude spot (in lambda_c/D)
   -rhoOuter: radius of outer opaque FPM ring (in lambda_c/D)
   -angDeg: angular opening (degrees) on the left/right/both sides.
   -whichSide: which sides to have open. 'left','right', 'top', 'bottom', or 'both'
   -centering: centering of the coordinates. 'pixel' or 'interpixel'
   -FOV: minimum desired field of view (in lambda_c/D)
   -shape: 'square' makes a square. Omitting makes a circle. 
   -clockAngDeg: Dark hole rotation about the z-axis (deg)

    Returns
    -------
    maskSW: rectangular, even-sized, binary-valued software mask
    xis: vector of coordinates along the horizontal axis (in lambda_c/D)
    etas: : vector of coordinates along the vertical axis (in lambda_c/D)
    """
    check.is_dict(inputs, 'inputs')

    # Required inputs
    pixresFP = inputs["pixresFP"]  # pixels per lambda_c/D
    rhoInner = inputs["rhoInner"]  # radius of inner FPM amplitude spot (in lambda_c/D)
    rhoOuter = inputs["rhoOuter"]  # radius of outer opaque FPM ring (in lambda_c/D)
    angDeg = inputs["angDeg"]  # angular opening (input in degrees)
    angRad = np.radians(angDeg)
    whichSide = inputs["whichSide"].lower()  # which (sides) of the dark hole have open
    
    # Optional inputs
    centering = inputs.get("centering", "pixel")
    check.centering(centering)
    darkHoleShape = inputs.get("shape", "circle").lower()
    clockAngDeg = inputs.get("clockAngDeg", 0)
    FOV = inputs.get("FOV", rhoOuter)
    xiOffset = inputs.get("xiOffset", 0.)
    etaOffset = inputs.get("etaOffset", 0.)
    if darkHoleShape in {'square', 'rect', 'rectangle'}:
        maxExtent = np.max((1, 2*np.abs(np.cos(np.radians(clockAngDeg)))))
    else:
        maxExtent = 1
    minFOVxi = inputs.get("xiFOV", maxExtent*FOV + np.abs(xiOffset))
    minFOVeta = inputs.get("etaFOV", maxExtent*FOV + np.abs(etaOffset))

    # Output array dimensions
    if centering == "pixel":
        Nxi0 = ceil_even(2*(minFOVxi*pixresFP + 1/2))
        Neta0 = ceil_even(2*(minFOVeta*pixresFP + 1/2))
    elif centering == "interpixel":
        Nxi0 = ceil_even(2*minFOVxi*pixresFP)
        Neta0 = ceil_even(2*minFOVeta*pixresFP)
    Nxi = inputs.get("Nxi", Nxi0)
    Neta = inputs.get("Neta", Neta0)

    # Focal Plane Coordinates
    deta = dxi = 1/pixresFP
    if centering == "interpixel":
        xis = np.arange(-(Nxi - 1)/2, (Nxi + 1)/2)*dxi
        etas = np.arange(-(Neta - 1)/2, (Neta + 1)/2)*deta
    elif centering == "pixel":
        xis = np.arange(-Nxi/2, Nxi/2) * dxi
        etas = np.arange(-Neta/2, Neta/2) * deta

    [XIS, ETAS] = np.meshgrid(xis, etas)
    XIS = XIS - xiOffset
    ETAS = ETAS - etaOffset
    RHOS = np.sqrt(XIS ** 2 + ETAS ** 2)
    THETAS = np.arctan2(ETAS, XIS)

    if whichSide in {'r', 'right', 'lr', 'rl', 'leftright', 'rightleft',
                     'both'}:
        clockAngRad = 0
    elif whichSide in {'l', 'left'}:
        clockAngRad = np.pi
    elif whichSide in {'t', 'top', 'u', 'up', 'tb', 'bt', 'ud', 'du',
                       'topbottom', 'bottomtop', 'updown', 'downup'}:
        clockAngRad = np.pi/2
    elif whichSide in {'b', 'bottom', 'd', 'down'}:
        clockAngRad = 3/2*np.pi
    else:
        raise ValueError('Invalid value given for inputs["whichSide"]')
    
    clockAngRad = clockAngRad + np.radians(clockAngDeg)

    # Generate the Outer Mask
    eps = np.finfo(float).eps
    rhoInner = rhoInner - 13*eps # Avoidy a ratty line from the higher numerical noise floor introduced by RHOS*cos().
    rhoOuter = rhoOuter + 13*eps # Avoidy a ratty line from the higher numerical noise floor introduced by RHOS*cos().
    if darkHoleShape in {'circle', 'annulus'}:
        softwareMask0 = np.logical_and(RHOS >= rhoInner,
                                       RHOS <= rhoOuter)
    elif darkHoleShape in {'square'}:
        softwareMask0 = np.logical_and(
            np.logical_or(
            np.logical_and(np.logical_and(np.logical_and(RHOS*cos(THETAS-clockAngRad)<=rhoOuter,
                                                         RHOS*cos(THETAS-clockAngRad)>=-rhoOuter),
                                                         RHOS*sin(THETAS-clockAngRad)<=rhoOuter),
                                                         RHOS*sin(THETAS-clockAngRad)>=-rhoOuter),
            np.logical_and(np.logical_and(np.logical_and(RHOS*cos(THETAS-clockAngRad)>=-rhoOuter,
                                                         RHOS*cos(THETAS-clockAngRad)<=rhoOuter),
                                                         RHOS*sin(THETAS-clockAngRad)<=rhoOuter),
                                                         RHOS*sin(THETAS-clockAngRad)>=-rhoOuter)
            ),
            RHOS >= rhoInner
        )
    elif darkHoleShape in {'rect', 'rectangle'}:
        softwareMask0 = np.logical_or(
            np.logical_and(np.logical_and(np.logical_and(RHOS*cos(THETAS-clockAngRad)>=rhoInner,
                                                         RHOS*cos(THETAS-clockAngRad)<=rhoOuter),
                                                         RHOS*sin(THETAS-clockAngRad)<=rhoOuter),
                                                         RHOS*sin(THETAS-clockAngRad)>=-rhoOuter),
            np.logical_and(np.logical_and(np.logical_and(RHOS*cos(THETAS-clockAngRad)<=-rhoInner,
                                                         RHOS*cos(THETAS-clockAngRad)>=-rhoOuter),
                                                         RHOS*sin(THETAS-clockAngRad)<=rhoOuter),
                                                         RHOS*sin(THETAS-clockAngRad)>=-rhoOuter)
            )
    elif darkHoleShape in {'d'}:
        softwareMask0 = np.logical_and(np.logical_or(
            RHOS*cos(THETAS-clockAngRad) >= rhoInner,
            RHOS*cos(THETAS-clockAngRad) <= -rhoInner),
            RHOS <= rhoOuter)
    else:
        raise ValueError('Invalid value given for inputs["shape"].')
    
    softwareMask = np.logical_and(softwareMask0, np.abs(np.angle(np.exp(1j*(THETAS-clockAngRad)))) <= angRad/2)
    
    if whichSide in {'both', 'lr', 'rl', 'leftright', 'rightleft', 'tb', 'bt',
                     'ud', 'du', 'topbottom', 'bottomtop', 'updown', 'downup'}:
        softwareMask2 = np.logical_and(softwareMask0, np.abs(np.angle(np.exp(1j *
                                (THETAS-(clockAngRad+np.pi))))) <= angRad/2)
        softwareMask = np.logical_or(softwareMask, softwareMask2)

    # # Generate the Outer Mask
    # # maskSW = 1.0 * (RHOS >= rhoInner) * (RHOS <= rhoOuter) * (THETAS <= angRad/2) * \
    # #     (THETAS >= -angRad/2)
    # # maskSW = 1.0 * np.logical_and(RHOS>=rhoInner,RHOS<=rhoOuter)
    # if DHshape in ('square'):
    #     maskSW0 = np.logical_and(RHO >= rhoInner, np.logical_and(abs(XIS) <= rhoOuter,
    #                                                              abs(ETAS) <= rhoOuter))
    # else:
    #     maskSW0 = np.logical_and(RHO >= rhoInner, RHO <= rhoOuter)

    # # If the user doesn't pass the clocking angle
    # if not hasattr(inputs, 'clockAngDeg'):
    #     if whichSide in ("L", "left"):
    #         clockAng = np.pi
    #     elif whichSide in ("R", "right"):
    #         clockAng = 0
    #     elif whichSide in ("T", "top"):
    #         clockAng = np.pi/2
    #     elif whichSide in ("B", "bottom"):
    #         clockAng = 3*np.pi/2
    #     else:
    #         clockAng = 0
    # else:
    #     clockAng = inputs["clockAngDeg"]*np.pi/180

    # maskSW = np.logical_and(maskSW0, np.abs(np.angle(np.exp(1j*(THETAS-clockAng)))) <= angRad/2)

    # if whichSide in ('both'):
    #     clockAng = clockAng + np.pi
    #     maskSW2 = np.logical_and(maskSW0, np.abs(np.angle(np.exp(1j*(THETAS-clockAng)))) <=
    #                              angRad/2)
    #     maskSW = np.logical_or(maskSW, maskSW2)

    return softwareMask, xis, etas


def falco_gen_bowtie_FPM(inputs):
    """
    Generate a bowtie FPM using PROPER.

    Outside the outer ring is opaque.If rhoOuter = infinity, then the outer
    ring is omitted and the mask is cropped down to the size of the inner spot.
    The inner spot has a specifyable amplitude value. The output array is the
    smallest size that fully contains the mask.

    Parameters
    ----------
     pixresFPM:  resolution in pixels per lambda_c/D
     rhoInner:   radius of inner FPM amplitude spot (in lambda_c/D)
     rhoOuter:   radius of outer opaque FPM ring (in lambda_c/D). Set to
                 infinity for an occulting-spot only FPM
     FPMampFac:  amplitude transmission of inner FPM spot
     centering:  pixel centering 

    Returns
    -------
     mask:    cropped-down, 2-D FPM representation. amplitude only 

    """
    check.is_dict(inputs, 'inputs')

    # Set default values of input parameters
    flagRot180deg = False

    ang = inputs["ang"]  # Opening angle on each side of the bowtie
    pixresFPM = inputs["pixresFPM"]
    rhoInner = inputs["rhoInner"]
    rhoOuter = inputs["rhoOuter"] 
    centering = inputs["centering"]
    FPMampFac = 0

    if hasattr(inputs, 'flagRot180deg'):
        flagRot180deg = inputs["flagRot180deg"]

    dxiUL = 1.0 / pixresFPM  # lambda_c/D per pixel. "UL" for unitless

    if centering == "interpixel":
        Narray = ceil_even(2 * rhoOuter / dxiUL)
    else:
        Narray = ceil_even(2 * (rhoOuter / dxiUL + 0.5))

    xshift = 0  # translation in x of FPM (in lambda_c/D)
    yshift = 0  # translation in y of FPM (in lambda_c/D)

    Darray = Narray * dxiUL  # width of array in lambda_c/D
    diam = Darray
    wl_dummy = 1e-6  # wavelength (m); Dummy value--no propagation here, so not used.

    if centering == "interpixel":
        cshift = -diam / 2 / Narray
    elif flagRot180deg:
        cshift = -diam / Narray
    else:
        cshift = 0

    # INITIALIZE PROPER. Note that:  bm.dx = diam / bdf / np;
    wf = proper.prop_begin(diam, wl_dummy, Narray, 1.0)

    # Outer opaque ring of FPM
    cx_OD = 0 + cshift + xshift
    cy_OD = 0 + cshift + yshift
    proper.prop_circular_aperture(wf, rhoOuter, cx_OD, cy_OD)

    # Inner spot of FPM (Amplitude transmission can be nonzero)
    cx_ID = 0 + cshift + xshift
    cy_ID = 0 + cshift + yshift
    innerSpot = proper.prop_ellipse(wf, rhoInner, rhoInner, cx_ID,
                                    cy_ID, DARK=True) * (1 - FPMampFac) + FPMampFac

    # Create the bowtie region
    if ang < 180:
        # Top part
        Lside = 2*rhoOuter
        xvert = cshift + xshift + np.array([0, Lside*cosd(ang/2), Lside*cosd(ang/2),
                                            -Lside*cosd(ang/2), -Lside*cosd(ang/2)])
        yvert = cshift + xshift + np.array([0, Lside*sind(ang/2), Lside, Lside,
                                            Lside*sind(ang/2)])
        bowtieTop = proper.prop_irregular_polygon(wf, xvert, yvert, DARK=True)

        # Bottom part
        xvert = cshift + xshift + np.array([0, Lside*cosd(ang/2), Lside*cosd(ang/2),
                                            -Lside*cosd(ang/2), -Lside*cosd(ang/2)])
        yvert = cshift + xshift + -1*np.array([0, Lside*sind(ang/2), Lside, Lside,
                                               Lside*sind(ang/2)])
        bowtieBottom = proper.prop_irregular_polygon(wf, xvert, yvert, DARK=True)
    else:
        bowtieTop = 1
        bowtieBottom = 1

    mask = np.fft.ifftshift(np.abs(wf.wfarr))  # undo PROPER's fftshift
    return mask * innerSpot * bowtieTop * bowtieBottom


def gen_annular_FPM(inputs):
    return falco_gen_annular_FPM(inputs)

def falco_gen_annular_FPM(inputs):
    """
    Generate an annular FPM using PROPER.

    Outside the outer ring is opaque.If rhoOuter = infinity, then the outer
    ring is omitted and the mask is cropped down to the size of the inner spot.
    The inner spot has a specifyable amplitude value. The output array is the
    smallest size that fully contains the mask.

    Parameters
    ----------
     pixresFPM:  resolution in pixels per lambda_c/D
     rhoInner:   radius of inner FPM amplitude spot (in lambda_c/D)
     rhoOuter:   radius of outer opaque FPM ring (in lambda_c/D). Set to
                 infinity for an occulting-spot only FPM
     FPMampFac:  amplitude transmission of inner FPM spot
     centering:  pixel centering

    Returns
    -------
     mask:    cropped-down, 2-D FPM representation. amplitude only

    """
    check.is_dict(inputs, 'inputs')

    # Set default values of input parameters
    flagRot180deg = False

    pixresFPM = inputs["pixresFPM"]
    rhoInner = inputs["rhoInner"]
    rhoOuter = inputs["rhoOuter"] 
    FPMampFac = inputs["FPMampFac"] 
    centering = inputs["centering"] 

    if hasattr(inputs, 'flagRot180deg'):
        flagRot180deg = inputs["flagRot180deg"]

    dxiUL = 1.0 / pixresFPM  # lambda_c/D per pixel. "UL" for unitless
    if np.isinf(rhoOuter):
        if centering == "interpixel":
            Narray = ceil_even((2 * rhoInner / dxiUL))
        else:
            Narray = ceil_even(2 * (rhoInner / dxiUL + 0.5))
    else:
        if centering == "interpixel":
            Narray = ceil_even(2 * rhoOuter / dxiUL)
        else:
            Narray = ceil_even(2 * (rhoOuter / dxiUL + 0.5))

    xshift = 0  # translation in x of FPM (in lambda_c/D)
    yshift = 0  # translation in y of FPM (in lambda_c/D)

    Darray = Narray * dxiUL  # width of array in lambda_c/D
    diam = Darray
    wl_dummy = 1e-6  # wavelength (m); Dummy value--no propagation here, so not used.

    if centering == "interpixel":
        cshift = -diam / 2 / Narray
    elif flagRot180deg:
        cshift = -diam / Narray
    else:
        cshift = 0

    # INITIALIZE PROPER. Note that:  bm.dx = diam / bdf / np;
    wf = proper.prop_begin(diam, wl_dummy, Narray, 1.0)

    if not np.isinf(rhoOuter):
        # Outer opaque ring of FPM
        cx_OD = 0 + cshift + xshift
        cy_OD = 0 + cshift + yshift
        proper.prop_circular_aperture(wf, rhoOuter, cx_OD, cy_OD)

    # Inner spot of FPM (Amplitude transmission can be nonzero)
    cx_ID = 0 + cshift + xshift
    cy_ID = 0 + cshift + yshift
    innerSpot = proper.prop_ellipse(wf, rhoInner, rhoInner, cx_ID,
                                    cy_ID, DARK=True) * (1 - FPMampFac) + FPMampFac

    mask = np.fft.ifftshift(np.abs(wf.wfarr))  # undo PROPER's fftshift
    return mask * innerSpot  # Include the inner FPM spot


def falco_gen_bowtie_LS(inputs):
    """
    Generate a sideways-bowtie-shaped Lyot stop using PROPER.

    Parameters
    ----------
    inputs : dict
        dictionary of input values.

    Returns
    -------
    mask : numpy ndarray
        2-D output mask.

    """
    check.is_dict(inputs, 'inputs')

    Nbeam = inputs["Nbeam"]  # number of points across the incoming beam
    ID = inputs["ID"]  # inner diameter of mask (in pupil diameters)
    OD = inputs["OD"]  # outer diameter of mask (in pupil diameters)
    ang = inputs["ang"]  # opening angle of the upper and lower bowtie wedges [degrees]

    Dbeam = 1.  # inputs.Dbeam; # diameter of the beam at the mask [pupil diameters]
    dx = Dbeam/Nbeam
    Dmask = Dbeam  # width of the beam (so can have zero padding if LS is undersized) [meters]

    # Optional inputs and their defaults
    centering = inputs['centering'] if 'centering' in inputs.keys() else 'pixel'  # Default to pixel centering
    xShear = inputs['xShear'] if 'xShear' in inputs.keys() else  0.  # x-axis shear of mask [pupil diameters]
    yShear = inputs['yShear'] if 'yShear' in inputs.keys() else  0.  # y-axis shear of mask [pupil diameters]
    clocking = -inputs['clocking'] if 'clocking' in inputs.keys() else 0.  # Clocking of the mask [degrees]
    magfac = inputs['magfac'] if 'magfac' in inputs.keys() else 1.  # magnification factor of the pupil diameter

    if(centering == 'pixel'):
        Narray = ceil_even(magfac*Nbeam + 1 + 2*Nbeam*np.max(np.abs(np.array([xShear, yShear]))))
    elif(centering == 'interpixel'):
        Narray = ceil_even(magfac*Nbeam + 2*Nbeam*np.max(np.abs(np.array([xShear, yShear]))))

    Darray = Narray*dx  # width of the output array [meters]
    bdf = Dmask/Darray  # beam diameter factor in output array
    wl_dummy = 1e-6  # wavelength (m); Dummy value--no propagation here, so not used.

    # No shift for pixel-centered pupil, or -Dbeam/Narray/2 shift for inter-pixel centering
    cshift = -dx/2 if 'interpixel' in centering else 0.

    # % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

    # INITIALIZE PROPER
    bm = proper.prop_begin(Dmask, wl_dummy, Narray, bdf)

    # PRIMARY MIRROR (OUTER DIAMETER)
    ra_OD = (Dbeam*OD/2.)*magfac
    cx_OD = cshift + xShear
    cy_OD = cshift + yShear
    proper.prop_circular_aperture(bm, ra_OD, cx_OD, cy_OD)

    # SECONDARY MIRROR (INNER DIAMETER)
    ra_ID = (Dbeam*ID/2.)*magfac
    cx_ID = cshift + xShear
    cy_ID = cshift + yShear
    proper.prop_circular_obscuration(bm, ra_ID, cx_ID, cy_ID)

    mask = np.fft.ifftshift(np.abs(bm.wfarr))

    # Create the bowtie region
    if(ang < 180):
        ang2 = 90. - ang/2.
        bm2 = bm
        Lside = 1.1*ra_OD  # Have the triangle go a little past the edge of the circle

        yvert0 = np.array([0., Lside*sind(ang2), Lside*sind(ang2),
                           -Lside*sind(ang2), -Lside*sind(ang2), 0.])

        # Right triangular obscuration
        xvert0 = np.array([0., Lside*cosd(ang2), Lside, Lside,
                           Lside*cosd(ang2), 0.])
        xvert = xvert0.copy()
        yvert = yvert0.copy()
        for ii in range(len(xvert0)):
            xy = np.array([[cosd(clocking), sind(clocking)],
                           [-sind(clocking), cosd(clocking)]]) @ \
                np.array([xvert0[ii], yvert0[ii]]).reshape(2, 1)
            xvert[ii] = xy[0]
            yvert[ii] = xy[1]
            pass
        bowtieRight = proper.prop_irregular_polygon(bm, cshift+xShear+xvert,
                                                    cshift+yShear+yvert, DARK=True)

        # Left triangular obscuration
        xvert0 = -np.array([0., Lside*cosd(ang2), Lside, Lside,
                            Lside*cosd(ang2), 0.])
        xvert = xvert0.copy()
        yvert = yvert0.copy()
        for ii in range(len(xvert0)):
            xy = np.array([[cosd(clocking), sind(clocking)],
                           [-sind(clocking), cosd(clocking)]]) @ \
                np.array([xvert0[ii], yvert0[ii]]).reshape(2, 1)
            xvert[ii] = xy[0]
            yvert[ii] = xy[1]
            pass
        bowtieLeft = proper.prop_irregular_polygon(bm2, cshift+xShear+xvert,
                                                   cshift+yShear+yvert, DARK=True)

        mask = mask*bowtieRight*bowtieLeft
        pass

    return mask


def falco_gen_pupil_LUVOIR_A_final(inputs, **kwargs):
    """
    Generate the LUVOIR Design A (Final) telescope pupil.

    Parameters
    ----------
    inputs : dict
        dictionary of input values.

    Returns
    -------
    mask : numpy ndarray
        2-D output mask.

    """
    check.is_dict(inputs, 'inputs')

    # Optional inputs and their defaults
    flagRot180deg = True if 'ROT180' in kwargs and kwargs["ROT180"] == True else False
    centering = inputs['centering'] if 'centering' in inputs.keys() else 'pixel'  # Default to pixel centering
    magfacD = inputs['magfacD'] if 'magfacD' in inputs.keys() else 1.0  # Default to no magnification of the pupil
    hexgap0 = inputs['wGap_m'] if 'wGap_m' in inputs.keys() else 6.0e-3  # Default to 6 mm between segments

    # % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

    # USER INPUTS
    Nbeam = inputs["Nbeam"]  # number of points across the incoming beam
    Nap = Nbeam  # number of points across FULL usable pupil
    width_hex0 = 1.2225  # flat-to-flat (m)
    Dap = (12*width_hex0 + 11*hexgap0)  # (12*width_hex0 + 12*hexgap0)
    dx = Dap/Nap

    if("pixel" == centering):
        Narray = ceil_even(Nbeam/magfacD+1)
        cshift = -dx if flagRot180deg else 0.
    elif("interpixel" == centering):
        Narray = ceil_even(Nbeam/magfacD)
        cshift = -dx/2.
    else:
        pass

    # For PROPER
    wl_dummy = 1e-6  # wavelength (m)
    bdf = Nbeam/Narray  # beam diameter factor in output array

    dx_t = 0.
    dy_t = 0.

    width_hex = magfacD*width_hex0  # flat-to-flat (m)
    nrings = 6
    hexrad = 2./np.sqrt(3.)*width_hex/2.
    hexgap = magfacD*hexgap0  # (m)
    hexsep = width_hex + hexgap  # distance from center to center of neighboring segments
    wStrut = 0.15*magfacD  # meters

    # Generate the input pupil for LUVOIR
    bm = proper.prop_begin(Dap, wl_dummy, Narray, bdf)

    # Subtract the inner ring from all the rings
    ap = falco_hex_aperture_LUVOIR_A(bm, nrings, hexrad, hexsep, cshift-dx_t,
                                     cshift-dy_t, DARK=True)

    # Add the struts
    proper.prop_rectangular_obscuration(bm, wStrut, 7*width_hex,
                                        cshift-dx_t, cshift-dy_t + magfacD*Dap/4.)

    len_1b = (np.sqrt(93)+0.5)*hexrad
    proper.prop_rectangular_obscuration(bm, wStrut, len_1b, cshift-dx_t + 1.5*hexrad,
                                        cshift-dy_t - 11*np.sqrt(3)/4*hexrad, ROTATION=12.7)
    proper.prop_rectangular_obscuration(bm, wStrut, len_1b, cshift-dx_t - 1.5*hexrad,
                                        cshift-dy_t - 11*np.sqrt(3)/4*hexrad, ROTATION=-12.7)

    mask = np.fft.ifftshift(np.abs(bm.wfarr))*ap
    mask[mask > 1] = 1.  # Get rid of overlapping segment edges at low res if gap size is 0.
    if(flagRot180deg):
        mask = np.rot90(mask, 2)

    return mask


def falco_gen_pupil_LUVOIR_A_final_Lyot(inputs, **kwargs):

    # Optional inputs and their defaults
    flagRot180deg = True if 'ROT180' in kwargs and kwargs["ROT180"]==True else False
    centering = inputs['centering'] if 'centering' in inputs.keys() else 'pixel'
    magfacD = inputs['magfacD'] if 'magfacD' in inputs.keys() else 1.0
    hexgap0 = inputs['wGap_m'] if 'wGap_m' in inputs.keys() else 6.0e-3

    # % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

    # USER INPUTS
    Nbeam = inputs["Nbeam"]  # number of points across the incoming beam  
    ID = inputs["ID"]  # inner diameter of mask (in pupil diameters)
    OD = inputs["OD"]  # outer diameter of mask (in pupil diameters)
    wStrut = inputs["wStrut"]  # width of the struts (in pupil diameters)
    wStrut = wStrut*Dbeam  # now in meters

    Nap = Nbeam  # number of points across FULL usable pupil
    width_hex0 = 1.2225  # flat-to-flat (m)
    Dap = (12*width_hex0 + 11*hexgap0)  # (12*width_hex0 + 12*hexgap0);
    dx = Dap/Nap
    dx_drawing = 1.2225/158

    if("pixel" == centering):
        Narray = ceil_even(Nbeam/magfacD+1)
        cshift = -dx if flagRot180deg else 0.
    elif("interpixel" == centering):
        Narray = ceil_even(Nbeam/magfacD)
        cshift = -dx/2.
    else:
        pass

    Darray = Narray*dx

    # For PROPER
    wl_dummy = 1e-6  # wavelength (m)
    bdf = Nbeam/Narray  # beam diameter factor in output array

    dx_t = 0.
    dy_t = 0.

    width_hex = magfacD*width_hex0  # flat-to-flat (m)
    hexrad = 2./np.sqrt(3.)*width_hex/2.
    wStrut = 0.15*magfacD  # meters

    # Generate the input pupil for LUVOIR
    bm = proper.prop_begin(Dap, wl_dummy, Narray, bdf)

    # Outer pupil edge
    proper.prop_circular_aperture(bm, OD/2, cshift-dx_t, cshift-dy_t)

    # Central Obscuration
    proper.prop_circular_obscuration(bm, ID/2., cshift-dx_t, cshift-dy_t)

    # Add the struts
    proper.prop_rectangular_obscuration(bm, wStrut, 7*width_hex, cshift-dx_t,
                                        cshift-dy_t + magfacD*Dap/4.)
    len_1b = (np.sqrt(93)+0.5)*hexrad
    proper.prop_rectangular_obscuration(bm, wStrut, len_1b, cshift-dx_t + 1.5*hexrad,
                                        cshift-dy_t - 11*np.sqrt(3)/4*hexrad, ROTATION=12.7)
    proper.prop_rectangular_obscuration(bm, wStrut, len_1b, cshift-dx_t - 1.5*hexrad,
                                        cshift-dy_t - 11*np.sqrt(3)/4*hexrad, ROTATION=-12.7)

    mask = np.fft.ifftshift(np.abs(bm.wfarr))
    if(flagRot180deg):
        mask = np.rot90(mask, 2)

    return mask


def falco_gen_pupil_LUVOIR_B_PROPER(inputs, **kwargs):
    """
    Generate the LUVOIR-B pupil with PROPER.

    Parameters
    ----------
    inputs : dict
        Dictionary of input parameters.
    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    mask : numpy ndarray

    """
    # Coordinates of hex segments to skip:
    # 1 13 114 115 126 127
    # 1 12 113 114 125 126

    # Optional inputs and their defaults
    centering = inputs['centering'] if 'centering' in inputs else 'pixel'
    magfacD = inputs['magfacD'] if 'magfacD' in inputs else 1.0
    hexgap0 = inputs['wGap_m'] if 'wGap_m' in inputs else 6.0e-3

    # % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

    # USER INPUTS
    Nbeam = inputs["Nbeam"]  # number of points across the incoming beam
    # Nap = ceil_even(1.1*Nbeam) # number of points across FULL usable pupil
    nrings = 4
    width_hex0 = 0.955  # flat-to-flat (m)
    Dap = ((2*nrings)*width_hex0 + (2*nrings-1)*hexgap0)  # (12*width_hex0 + 12*hexgap0)
    dx = Dap/Nbeam
    # dx_drawing = 1.2225/158 # (m)

    if("pixel" == centering):
        Narray = ceil_even(1.02*Nbeam/magfacD+1)
        cshift = 0.
    elif("interpixel" == centering):
        Narray = ceil_even(1.02*Nbeam/magfacD)
        cshift = -dx/2.
    else:
        pass

    # For PROPER
    wl_dummy = 1e-6  # wavelength (m)
    bdf = Nbeam/Narray  # beam diameter factor in output array

    dx_t = 0.
    dy_t = 0.

    width_hex = magfacD*width_hex0  # flat-to-flat (m)

    hexrad = 2./np.sqrt(3.)*width_hex/2.
    hexgap = magfacD*hexgap0  # (m)
    hexsep = width_hex + hexgap  # segments' center-to-center distance

    # Generate the input pupil for LUVOIR
    bm = proper.prop_begin((Narray/Nbeam)*Dap, wl_dummy, Narray, bdf)
    ap = falco_hex_aperture_LUVOIR_B(bm, nrings, hexrad, hexsep, cshift-dx_t, cshift-dy_t)
    mask = np.fft.ifftshift(np.abs(bm.wfarr))*ap

    return mask


def falco_hex_aperture_LUVOIR_A(wf, nrings, hexrad, hexsep, xc=0.0, yc=0.0, **kwargs):
    """
    Return a mask consisting of multiple hexagons for the LUVOIR A pupil.

    This is used for generating the primary mirror for the LUVOIR telescope.
    The hexagons have antialiased edges. This routine does not modify the
    wavefront.

    Parameters
    ----------
    wf : object
        WaveFront class object
    nrings : int
        Number of rings of hexagons in aperture (e.g. 1 = a central hexagon
        surrounded by a ring of hexagons)
    hexrad : float
        Distance in meters from center of a hexagon segment to a vertex.
    hexsep : float
        The distance between the centers of adjacent hexagons.
    xc, yc : float
        The offset in meters of the aperture from the center of the wavefront.
        By default, the aperture is centered within the wavefront.

    Optional Keywords
    -----------------
    DARK : boolean
        If set, the central hexagonal segment will be set to 0.0.
    ROTATION : float
        The counterclockwise rotation in degrees of the aperture about its center.

    Returns
    -------
       ap : numpy ndarray
        2-D mask made of hexagons
    """
    ngrid = wf.ngrid

    ap = np.zeros([ngrid, ngrid], dtype=np.float64)

    if "ROTATION" in kwargs:
        angle = kwargs["ROTATION"]
        angle_rad = angle * np.pi/180.
    else:
        angle = 0.0
        angle_rad = 0.0

    counter = 0
    for iring in range(0, nrings+1):
        x = hexsep * np.cos(30 * np.pi/180.) * iring
        y = -nrings * hexsep + iring * hexsep * 0.5
        for iseg in range(0, 2*nrings-iring+1):
            xhex = x * np.cos(angle_rad) - y * np.sin(angle_rad) + xc
            yhex = x * np.sin(angle_rad) + y * np.cos(angle_rad) + yc
            if (iring != 0 or not (iseg == nrings and "DARK" in kwargs)):
                counter += 1
                if not any(counter == np.array([1, 12, 113, 114, 125, 126])):
                    ap = ap + proper.prop_polygon(wf, 6, hexrad, xhex, yhex,
                                                  rotation=angle)

            if (iring != 0):
                xhex = -x * np.cos(angle_rad) - y * np.sin(angle_rad) + xc
                yhex = -x * np.sin(angle_rad) + y * np.cos(angle_rad) + yc
                counter += 1
                if not any(counter == np.array([1, 12, 113, 114, 125, 126])):
                    ap = ap + proper.prop_polygon(wf, 6, hexrad, xhex, yhex,
                                                  rotation=angle)

            y += hexsep

    return ap


def falco_hex_aperture_LUVOIR_B(wf, nrings, hexrad, hexsep, xc=0., yc=0., **kwargs):
    """
    Return a mask consisting of multiple hexagons for the LUVOIR B pupil.

    This is used for generating the primary mirror for the LUVOIR telescope. 
    The hexagons have antialiased edges. This routine does not modify the
    wavefront.

    Parameters
    ----------
    wf : object
        WaveFront class object
    nrings : int
        Number of rings of hexagons in aperture (e.g. 1 = a central hexagon
        surrounded by a ring of hexagons)
    hexrad : float
        The distance in meters from the center of a hexagon segment to a vertex.
    hexsep : float
        The distance between the centers of adjacent hexagons.
    xc, yc : float
        The offset in meters of the aperture from the center of the wavefront.
        By default, the aperture is centered within the wavefront.

    Optional Keywords
    -----------------
    DARK : boolean
        If set, the central hexagonal segment will be set to 0.0.
    ROTATION : float
        The counterclockwise rotation in degrees of the aperture about its center.

    Returns
    -------
        numpy ndarray
        A hexagonal mask
    """
    ngrid = wf.ngrid

    ap = np.zeros([ngrid, ngrid], dtype=np.float64)

    if "ROTATION" in kwargs:
        angle = kwargs["ROTATION"]
        angle_rad = angle * np.pi/180.
    else:
        angle = 0.0
        angle_rad = 0.0

    counter = 0
    for iring in range(0, nrings+1):
        x = hexsep * np.cos(30 * np.pi/180.) * iring
        y = -nrings * hexsep + iring * hexsep * 0.5
        for iseg in range(0, 2*nrings-iring+1):
            xhex = x * np.cos(angle_rad) - y * np.sin(angle_rad) + xc
            yhex = x * np.sin(angle_rad) + y * np.cos(angle_rad) + yc
            if (iring != 0 or not (iseg == nrings and "DARK" in kwargs)):
                counter += 1
                if not any(counter == np.array([1, 9, 52, 60])):
                    ap = ap + proper.prop_polygon(wf, 6, hexrad, xhex, yhex, rotation=angle)

            if (iring != 0):
                xhex = -x * np.cos(angle_rad) - y * np.sin(angle_rad) + xc
                yhex = -x * np.sin(angle_rad) + y * np.cos(angle_rad) + yc
                counter += 1
                if not any(counter == np.array([1, 9, 53, 61])):
                    ap = ap + proper.prop_polygon(wf, 6, hexrad, xhex, yhex, rotation=angle)

            y += hexsep

    return ap


def falco_gen_pupil_Simple(inputs):
    """
    Generate a custom simple circular pupil with an ID, OD, and struts.

    Parameters
    ----------
    inputs : dict
        Dictionary of input parameters

    Returns
    -------
    pupil : numpy ndarray
        2-D pupil mask

    """
    check.is_dict(inputs, 'inputs')

    # Required dictionary keys
    Nbeam = inputs["Nbeam"]  # Aperture diameter in pixel widths
    Narray = inputs["Npad"]  # Number of points across 2-D, NxN output array
    OD = inputs["OD"]  # pupil outer diameter, can be < 1

    # Optional dictionary keys
    wStrut = inputs.get("wStrut", 0.)  # width of each strut [pupil diameters]
    angStrut = inputs.get("angStrut", [])
    angStrutVec = np.atleast_1d(angStrut)
    
    # if 'angStrut' in inputs:
    #     angStrutVec = inputs["angStrut"]  # Azimuthal locations
    #     angStrutVec = np.array(angStrutVec)
    # else:
    #     angStrutVec = np.array([])
    #     wStrut = 0
    ID = inputs.get("ID", 0.)  # central obscuration diam [pupil diameters]
    centering = inputs.get("centering", "pixel")
    xStretch = inputs.get("xStretch", 1.)
    clocking = inputs.get("clocking", 0.)  # [degrees]
    xShear = inputs.get("xShear", 0.)  # [pupil diameters]
    yShear = inputs.get("yShear", 0.)  # [pupil diameters]
    flagHG = inputs.get("flagHG", False)
    
    # Checks on dict keys
    check.real_nonnegative_scalar(wStrut, 'wStrut', TypeError)
    check.centering(centering)
    if not isinstance(flagHG, bool):
        raise TypeError("inputs['flagHG'] must be a bool")
    if ID > OD:
        raise ValueError("Inner diameter is larger than outer diameter.")

    # By default, don't use hyger-gaussians for anti-aliasing the edges.
    if not flagHG:

        # Create outer aperture
        inpOuter = {}
        inpOuter["Nbeam"] = Nbeam
        inpOuter["Narray"] = Narray
        inpOuter["radiusX"] = xStretch*0.5*OD
        inpOuter["radiusY"] = 0.5*OD
        inpOuter["centering"] = centering
        inpOuter["clockingDegrees"] = clocking
        inpOuter["xShear"] = xShear
        inpOuter["yShear"] = yShear
        apOuter = gen_ellipse(inpOuter)

        # Create inner obscuration
        if ID > 0:
            inpInner = {}
            inpInner["Nbeam"] = Nbeam
            inpInner["Narray"] = Narray
            inpInner["radiusX"] = xStretch*0.5*ID
            inpInner["radiusY"] = 0.5*ID
            inpInner["centering"] = centering
            inpInner["clockingDegrees"] = clocking
            inpInner["xShear"] = xShear
            inpInner["yShear"] = yShear
            apInner = 1.0 - gen_ellipse(inpInner)
        else:
            apInner = 1

        # Create strut obscurations
        if angStrutVec.size == 0:
            apStruts = 1
        else:
            # INITIALIZE PROPER
            Dbeam = 1.0  # diameter of beam (normalized to itself)
            dx = Dbeam/Nbeam
            Darray = Narray*dx
            wl_dummy = 1e-6  # dummy value
            bdf = Dbeam/Darray  # beam diameter fraction
            if centering == 'pixel':
                cshift = 0
            elif centering == 'interpixel':
                cshift = -dx/2.
            bm = proper.prop_begin(Dbeam, wl_dummy, Narray, bdf)
            
            # STRUTS
            lStrut = 0.6  # [pupil diameters]
            rcStrut0 = lStrut / 2.0
            for iStrut in range(angStrutVec.size):
                ang = angStrutVec[iStrut] + clocking
                proper.prop_rectangular_obscuration(bm, lStrut, wStrut,
                                        rcStrut0*cosd(ang)+cshift+xShear,
                                        rcStrut0*sind(ang)+cshift+yShear,
                                        ROTATION=ang)
            apStruts = np.fft.ifftshift(np.abs(bm.wfarr))
        
        # Combine all features
        pupil = apOuter*apInner*apStruts
    
    else:
        hg_expon = 1000  # hyper-gaussian exponent for anti-aliasing
        hg_expon_spider = 100  # hyper-gaussian exponent for anti-aliasing
        apRad = Nbeam/2.  # aperture radius in samples
        
        # Create coordinates
        if centering == 'pixel':
            x = np.arange(-Narray/2, Narray/2)
        elif centering == 'interpixel':
            x = np.arange(-(Narray-1)/2, (Narray-1)/2+1)
        RHO = falco.util.radial_grid(x, xStretch=xStretch)
        THETA = falco.util.azimuthal_grid(x, xStretch=xStretch)
    
        if ID > 0:
            pupil = np.exp(-(RHO/(apRad*OD))**hg_expon) - \
                np.exp(-(RHO/(apRad*ID))**hg_expon)
        else:
            pupil = np.exp(-(RHO/(apRad*OD))**hg_expon)
            
        # Create spiders
        if wStrut > 0:
            try:
                halfwidth = wStrut*apRad
                for ang in angStrutVec:
                    pupil *= (1 - np.exp(-(RHO*np.sin(THETA-ang*np.pi/180)
                                           / halfwidth)**hg_expon_spider) *
                              (RHO*np.cos(THETA-ang*np.pi/180) > 0))
            except:
                raise TypeError("inputs['angStrut'] must be an iterable")

    return pupil


def falco_gen_pupil_customHex(inputs):
    """
    Generate a custom segmented pupil comprised of hexagonal segments.

    Parameters
    ----------
    inputs : dict
        Dictionary of input parameters

    Returns
    -------
    pupil : numpy ndarray
        2-D pupil mask

    """
    check.is_dict(inputs, 'inputs')

    hg_expon = 1000  # hyper-gaussian exponent for anti-aliasing
    hg_expon_spider = 100  # hyper-gaussian exponent for anti-aliasing

    N = inputs["Npad"]  # Number of samples in NxN grid
    OD = inputs["OD"]  # pupil outer diameter, can be < 1
    ID = inputs["ID"]  # central obscuration radius
    apRad = inputs["Nbeam"]/2.  # aperture radius in samples

    if 'wStrut' in inputs:
        wStrut = inputs["wStrut"]  # width of all struts [pupil diameters], float
        check.real_nonnegative_scalar(wStrut, 'wStrut', TypeError)

        angStrutVec = inputs["angStrut"]  # Azimuthal locations of the radial struts, array_like
    else:
        wStrut = 0

    # Create coordinates
    [X, Y] = np.meshgrid(np.arange(-N/2, N/2), np.arange(-N/2, N/2))
    [THETA, RHO] = falco.util.cart2pol(X, Y)

    inputs["apDia"] = inputs["Nbeam"]
    if('pistons' in inputs.keys()):
        pupil0 = falco.hexsegmirror.get_field(inputs)
    else:
        pupil0 = falco.hexsegmirror.get_support(inputs)

    # Create inner and outer circles
    if(ID > 0):
        pupil = np.exp(-(RHO/(apRad*OD))**hg_expon) - np.exp(-(RHO/(apRad*ID))**hg_expon)
    else:
        pupil = np.exp(-(RHO/(apRad*OD))**hg_expon)

    pupil = pupil*pupil0

    # Create spiders
    if(wStrut > 0):
        halfwidth = wStrut*2.*apRad
        try:
            for ang in angStrutVec:
                pupil = pupil*(1.-np.exp(-(RHO*np.sin(THETA-ang*np.pi/180.)/halfwidth)**hg_expon_spider) *
                                (RHO*np.cos(THETA-ang*np.pi/180.)>0))
        except:  # not iterable
            raise TypeError("inputs['angStrut'] must be an iterable")

    return pupil


def falco_gen_pupil_LUVOIR_B(Nbeam):
    """
    Generate the LUVOIR B pupil.

    Parameters
    ----------
    Nbeam : float, int
        Number of points across the pupil diameter.

    Returns
    -------
    pupil : numpy ndarray
        2-D pupil amplitude for LUVOIR B

    """
    D = 7.989  # meters, circumscribed. Segments 0.955m flat-to-flat. 6mm gaps.
    wGap = 6e-3/D  # Fractional width of segment gaps

    inputs = {}
    inputs["Nbeam"] = Nbeam/0.925  # number of points across the pupil diameter
    inputs["wGap"] = wGap*Nbeam  # number of samples across segment gaps
    inputs["numRings"] = 4  # Number of rings in hexagonally segmented mirror
    inputs["Npad"] = int(2**(falco.util.nextpow2(Nbeam)))
    inputs["ID"] = 0  # central obscuration radius
    inputs["OD"] = 1  # pupil outer diameter, can be < 1
    inputs["angStrut"] = np.array([])  # Angles of the struts (deg)
    inputs["wStrut"] = 0  # np.array([]) # Width of the struts (fraction of pupil diam.)

    missingSegments = np.ones(falco.hexsegmirror.count_segments(inputs["numRings"]),)
    for index in range(6):
        missingSegments[38+index*4 - 1] = 0

    inputs["missingSegments"] = missingSegments

    pupil = falco_gen_pupil_customHex(inputs)

    return pupil


def falco_gen_vortex_mask(charge, N):
    """
    Generate a vortex phase mask.

    Parameters
    ----------
    charge : int, float
        Charge of the vortex mask.
    N : int
        Number of points across output array.

    Returns
    -------
    vortex : numpy ndarray
        2-D vortex phase mask

    """
    check.real_scalar(charge, 'charge', TypeError)
    check.positive_scalar_integer(N, 'N', TypeError)
    return np.exp(1j*charge*falco.util.azimuthal_grid(np.arange(-N/2., N/2.)))


def gen_ellipse(inputs):
    
    pupil = falco_gen_ellipse(inputs)
    
    return pupil


def falco_gen_ellipse(inputs):
    """
    Generate a rotated ellipse with antialiased edges.

    Parameters
    ----------
    inputs : dict
        dictionary of input values.

    Returns
    -------
    pupil : numpy ndarray
        2-D output mask.

    """
    check.is_dict(inputs, 'inputs')

    Nbeam = inputs['Nbeam']
    Narray = inputs['Narray']
    radiusX = inputs['radiusX']
    radiusY = inputs['radiusY']

    # Optional dictionary keys
    centering = inputs["centering"] if('centering' in inputs) else 'pixel'
    clockingDegrees = inputs["clockingDegrees"] if('clockingDegrees' in inputs) else 0.
    clockingRadians = (np.pi/180.)*clockingDegrees
    xShear = inputs["xShear"] if('xShear' in inputs) else 0.
    yShear = inputs["yShear"] if('yShear' in inputs) else 0.
    magFac = inputs["magFac"] if('magFac' in inputs) else 1

    if centering == 'pixel':
        x = np.linspace(-Narray/2., Narray/2. - 1, Narray)/float(Nbeam)
    elif centering == 'interpixel':
        x = np.linspace(-(Narray-1)/2., (Narray-1)/2., Narray)/float(Nbeam)

    y = x
    x = x - xShear
    y = y - yShear
    [X, Y] = np.meshgrid(x,y)
    dx = x[1] - x[0]
    radius = 0.5

    RHO = 1/magFac*0.5*np.sqrt(
        1/(radiusX)**2*(np.cos(clockingRadians)*X + np.sin(clockingRadians)*Y)**2
        + 1/(radiusY)**2*(np.sin(clockingRadians)*X - np.cos(clockingRadians)*Y)**2
        )

    halfWindowWidth = np.max(np.abs((RHO[1, 0]-RHO[0, 0], RHO[0, 1] - RHO[0, 0])))
    pupil = -1*np.ones(RHO.shape)
    pupil[np.abs(RHO) < radius - halfWindowWidth] = 1
    pupil[np.abs(RHO) > radius + halfWindowWidth] = 0
    grayInds = np.array(np.nonzero(pupil == -1))
    # print('Number of grayscale points = %d' % grayInds.shape[1])

    upsampleFactor = 100
    dxUp = dx/float(upsampleFactor)
    xUp = np.linspace(-(upsampleFactor-1)/2., (upsampleFactor-1)/2., upsampleFactor)*dxUp
    # xUp = (-(upsampleFactor-1)/2:(upsampleFactor-1)/2)*dxUp
    [Xup, Yup] = np.meshgrid(xUp, xUp)

    subpixel = np.zeros((upsampleFactor, upsampleFactor))

    for iInterior in range(grayInds.shape[1]):

        subpixel = 0*subpixel

        xCenter = X[grayInds[0, iInterior], grayInds[1, iInterior]]
        yCenter = Y[grayInds[0, iInterior], grayInds[1, iInterior]]
        RHOup = 0.5*np.sqrt(
        1/(radiusX)**2*(np.cos(clockingRadians)*(Xup+xCenter) +
                        np.sin(clockingRadians)*(Yup+yCenter))**2
        + 1/(radiusY)**2*(np.sin(clockingRadians)*(Xup+xCenter) -
                          np.cos(clockingRadians)*(Yup+yCenter))**2)

        subpixel[RHOup <= radius] = 1
        pixelValue = np.sum(subpixel)/float(upsampleFactor**2)
        pupil[grayInds[0, iInterior], grayInds[1, iInterior]] = pixelValue

    return pupil


def rotate_shift_downsample_pupil_mask(arrayIn, nBeamIn, nBeamOut, xOffset,
                                       yOffset, rotDeg):
    """
    Translate, rotate, and downsample a pixel-centered mask.

    Parameters
    ----------
    arrayIn : np.ndarray
        2-D, pixel-centered array containing the pupil mask.
    nBeamIn : float
        Number of points across beam at starting resolution.
    nBeamOut : float
        Number of points across beam at final resolution.
    xOffset, yOffset : float
        x- and y-offsets of the mask in output-sized pixels
    rotDeg : float
        Amount to rotate the array about the center pixel in degrees.

    Returns
    -------
    arrayOut : np.ndarray
        2-D, even-sized, square array containing the resized mask.
    """
    check.twoD_array(arrayIn, 'arrayIn', TypeError)
    check.real_positive_scalar(nBeamIn, 'nBeamIn', TypeError)
    check.real_positive_scalar(nBeamOut, 'nBeamOut', TypeError)
    check.real_scalar(xOffset, 'xOffset', TypeError)
    check.real_scalar(yOffset, 'yOffset', TypeError)
    check.real_scalar(rotDeg, 'rotDeg', TypeError)

    if nBeamIn <= nBeamOut:
        raise ValueError('This function is for downsampling only.')
    else:
        if arrayIn.shape[0] % 2 == 0:  # if in an even-sized array
            # Crop assuming SP is pixel centered with row 0 and column 0 empty.
            arrayIn = arrayIn[1::, 1::]

        # Array sizes
        dxIn = 1./nBeamIn
        dxOut = 1./nBeamOut
        nArrayIn = arrayIn.shape[0]
        nArrayOut = falco.util.ceil_odd(nArrayIn*dxIn/dxOut + 2
                             + 2.*np.max((np.abs(xOffset), np.abs(yOffset))))
        # 2 pixels added to guarantee the offset mask is fully contained in
        # the output array.

        # array-centered coordinates of input matrix [pupil diameters]
        x0 = np.arange(-(nArrayIn-1.)/2., (nArrayIn)/2., 1)*dxIn
        [X0, Y0] = np.meshgrid(x0, x0)
        R0 = np.sqrt(X0**2 + Y0**2)
        Window = 0*R0
        Window[R0 <= dxOut/2.] = 1
        Window = Window/np.sum(Window)

        # To get good grayscale edges, convolve with the correct window
        # before downsampling.
        f_window = np.fft.ifft2(np.fft.ifftshift(Window))*nArrayIn
        f_arrayin = np.fft.ifft2(np.fft.ifftshift(arrayIn))*nArrayIn
        A = np.fft.fftshift(np.fft.fft2(f_window*f_arrayin))
        A = np.real(A)

        if not rotDeg == 0:
            A = rotate(A, -rotDeg, reshape=False)

        x1 = (np.arange(-(nArrayOut-1.)/2., nArrayOut/2., 1) - xOffset)*dxOut
        y1 = (np.arange(-(nArrayOut-1.)/2., nArrayOut/2., 1) - yOffset)*dxOut

        # RectBivariateSpline is faster in 2-D than interp2d
        interp_spline = RectBivariateSpline(x0, x0, A)
        Atemp = interp_spline(y1, x1)

        arrayOut = np.zeros((nArrayOut+1, nArrayOut+1))
        arrayOut[1::, 1::] = np.real(Atemp)

        return arrayOut



import os
INFLUENCE_XINETICS = os.path.dirname(os.path.abspath(__file__)) + "/data/influence_dm5v2.fits"
INFLUENCE_BMC_KILO = os.path.dirname(os.path.abspath(__file__)) + "/data/influence_BMC_kiloDM_300micron_res10_spline.fits"
INFLUENCE_BMC_2K = os.path.dirname(os.path.abspath(__file__)) + "/data/influence_BMC_2kDM_400micron_res10.fits"




"""Functions for setting up and generating HLC occulters."""

import numpy as np
from math import radians, sin, cos
from astropy.io import fits
from scipy.interpolate import griddata, RectBivariateSpline
import matplotlib.pyplot as plt

import falco
from falco.util import ceil_even, pad_crop, cart2pol
from falco.check import real_positive_scalar


def gen_fpm_from_LUT(mp, si, wi, modelType):
    """
    Produce the complex transmission of an HLC focal plane mask.

    Parameters
    ----------
    mp : TYPE
        DESCRIPTION.
    si : TYPE
        DESCRIPTION.
    wi : TYPE
        DESCRIPTION.
    modelType : {'compact', 'full'}
        String denoting whether to use the output is for the compact or full
        optical model.

    Returns
    -------
    fpm : numpy ndarray
        2-D, complex-valued representation of the HLC FPM.

    Notes
    -----
    In Matlab version is called falco_gen_fpm_from_LUT.
    """
    # Different variables for compact and full models
    if modelType.lower() == 'compact':
        # size = (Ndiel, Nmetal, mp.Nsbp)
        complexTransCube = mp.complexTransCompact
        # Index of the wavelength in mp.sbp_centers
        ilam = si
    elif modelType.lower() == 'full':
        # size = (Ndiel,Nmetal,mp.Nsbp*mp.Nwpsbp)
        complexTransCube = mp.complexTransFull
        # Index of the wavelength in mp.lam_array
        ilam = (si-1)*mp.Nwpsbp + wi
    else:
        raise ValueError('modelType must specify full or compact model.')

    t_diel_bias = mp.t_diel_bias_nm*1e-9  # bias thickness of dielectric [m]

    # Generate thickness profiles of each layer
    # Metal layer profile [m]
    DM8surf = falco.hlc.gen_fpm_surf_from_cube(mp.dm8, modelType)
    # Dielectric layer profile [m]
    DM9surf = t_diel_bias + falco.hlc.gen_fpm_surf_from_cube(mp.dm9, modelType)
    Nxi = DM8surf.shape[1]
    Neta = DM8surf.shape[0]

    if not (DM8surf.size == DM9surf.size):
        raise ValueError('FPM surf arrays for DM 8 and 9 must be same size!')

    # Obtain indices of nearest thickness values in the complex tran datacube.
    DM8transInd = falco.hlc.discretize_fpm_surf(DM8surf, mp.t_metal_nm_vec,
                                      mp.dt_metal_nm)
    DM9transInd = falco.hlc.discretize_fpm_surf(DM9surf, mp.t_diel_nm_vec,
                                      mp.dt_diel_nm)

    # Look up values
    fpm = np.zeros((Neta, Nxi), dtype=complex)
    for ix in range(Nxi):
        for iy in range(Neta):
            ind_metal = DM8transInd[iy, ix]
            ind_diel = DM9transInd[iy, ix]
            fpm[iy, ix] = complexTransCube[ind_diel, ind_metal, ilam]

    return fpm


def gen_fpm_cube_from_LUT(mp, modelType):
    """
    Produce the complex transmission of an HLC focal plane mask.

    Parameters
    ----------
    mp : TYPE
        DESCRIPTION.
    modelType : {'compact', 'full'}
        String denoting whether to use the output is for the compact or full
        optical model.

    Returns
    -------
    fpmCube : numpy ndarray
        Stack of 2-D, complex-valued HLC FPMs at each wavelength.

    Notes
    -----
    In Matlab version is called falco_gen_HLC_FPM_complex_trans_cube.
    """
    # Different variables for compact and full models
    if modelType.lower() == 'compact':
        # size = (Ndiel, Nmetal, mp.Nsbp)
        complexTransCube = mp.complexTransCompact
        lamVec = mp.sbp_centers
    elif modelType.lower() == 'full':
        # size = (Ndiel,Nmetal,mp.Nsbp*mp.Nwpsbp)
        complexTransCube = mp.complexTransFull
        lamVec = mp.full.lambdas
    else:
        raise ValueError('modelType must specify full or compact model.')
    Nlam = len(lamVec)

    t_diel_bias = mp.t_diel_bias_nm*1e-9  # bias thickness of dielectric [m]

    # Generate thickness profiles of each layer
    # Metal layer profile [m]
    DM8surf = falco.hlc.gen_fpm_surf_from_cube(mp.dm8, modelType)
    # Dielectric layer profile [m]
    DM9surf = t_diel_bias + falco.hlc.gen_fpm_surf_from_cube(mp.dm9,
                                                                 modelType)
    Neta, Nxi = DM8surf.shape

    if not (DM8surf.shape == DM9surf.shape):
        raise ValueError('FPM surf arrays for DM 8 and 9 must be same size!')

    # Obtain indices of nearest thickness values in the complex tran datacube.
    DM8transInd = falco.hlc.discretize_fpm_surf(DM8surf, mp.t_metal_nm_vec,
                                                mp.dt_metal_nm)
    DM9transInd = falco.hlc.discretize_fpm_surf(DM9surf, mp.t_diel_nm_vec,
                                                mp.dt_diel_nm)

    # Look up values
    fpmCube = np.zeros((Neta, Nxi, Nlam), dtype=complex)
    for ilam in range(Nlam):
        for ix in range(Nxi):
            for iy in range(Neta):
                ind_metal = DM8transInd[iy, ix]
                ind_diel = DM9transInd[iy, ix]
                fpmCube[iy, ix, ilam] = complexTransCube[ind_diel, ind_metal,
                                                         ilam]

    return fpmCube, DM8surf, DM9surf


def discretize_fpm_surf(FPMsurf, t_nm_vec, dt_nm):
    """
    Discretize the surface profiles of the FPM materials.

    Parameters
    ----------
    FPMsurf : numpy ndarray
        2-D array of the FPM material layer
    t_nm_vec : numpy ndarray
        The vector of allowed thickness values for the material layer
    dt_nm : float
        Scalar value of the step size between allowed thickness values
    
    Returns
    -------
    DMtransInd : numpy ndarray, dtype = int
        Array index of the metal/dielectric layer for each DM8/9 surface value
        in the complex transmission matrix.

    """
    # Convert surface profiles from meters to nanometers
    FPMsurf = 1e9*FPMsurf

    # Stay within the thickness range since material properties are not defined
    # outside it
    FPMsurf[FPMsurf < np.min(t_nm_vec)] = np.min(t_nm_vec)
    FPMsurf[FPMsurf > np.max(t_nm_vec)] = np.max(t_nm_vec)

    # Discretize to find the index the the complex transmission array
    DMtransInd = 0 + np.round(1/dt_nm*(FPMsurf - np.min(t_nm_vec)))

    return DMtransInd.astype(int)


def gen_fpm_surf_from_cube(dm, modelType):
    """
    Produce a FPM surface (in meters) with linear superposition.

    Function to produce a FPM surface (in meters). Uses linear superposition
    of a datacube of influence functions.

    Parameters
    ----------
    dm : TYPE
        DESCRIPTION.
    modelType : TYPE
        DESCRIPTION.

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    fpmSurf : TYPE
        DESCRIPTION.

    Notes
    -----
    In Matlab version is called falco_gen_HLC_FPM_surf_from_cube.

    """
    if modelType.lower() == 'compact':
        dmFullOrCompact = dm.compact
    elif modelType.lower() == 'full':
        dmFullOrCompact = dm
    else:
        raise ValueError('modelType must be full or compact.')

    fpmSurf = np.zeros((dmFullOrCompact.NdmPad, dmFullOrCompact.NdmPad))
    for iact in range(dm.NactTotal):
        if np.sum(np.abs(dmFullOrCompact.inf_datacube[:, :, iact])) >= 1e-8 \
            and np.sum(np.abs(dm.VtoH.flatten()[iact])) >= 1e-12:
            x_box_ind = np.arange(dmFullOrCompact.xy_box_lowerLeft[0, iact],
                                  dmFullOrCompact.xy_box_lowerLeft[0, iact]
                                  + dmFullOrCompact.Nbox, dtype=np.int)
            y_box_ind = np.arange(dmFullOrCompact.xy_box_lowerLeft[1, iact],
                                  dmFullOrCompact.xy_box_lowerLeft[1, iact]
                                  + dmFullOrCompact.Nbox, dtype=np.int)
            fpmSurf[np.ix_(y_box_ind, x_box_ind)] += dm.V.flatten()[iact] *\
            dm.VtoH.flatten()[iact]*dmFullOrCompact.inf_datacube[:, :, iact]

    # Non-negative values only
    fpmSurf[fpmSurf < 0] = 0.

    return fpmSurf


"""-------------------------------------------------------------------------"""


def setup_fpm_cosine(mp):
    
    if mp.F3.full.res != mp.F3.compact.res:
        raise ValueError('Resolution at F3 must be same for cosine basis set.')
    
    # Centering of DM surfaces on array
    mp.dm8.centering = mp.centering
    mp.dm9.centering = mp.centering

    mp.dm9.compact = mp.dm9
    
    mp.dm9.dxi = (mp.fl*mp.lambda0/mp.P2.D)/mp.F3.full.res  # width of a pixel at the FPM in the full model (meters)
    mp.dm9.compact.dxi = (mp.fl*mp.lambda0/mp.P2.D)/mp.F3.compact.res  # width of a pixel at the FPM in the compact model (meters)

    drCos = 1/mp.dm9.actres  # Width and double-separation of the cosine rings [lambda0/D]
    Nrad = int(np.ceil(2*mp.dm9.actres*mp.F3.Rin))

    # Generate datacube of influence functions, which are rings with radial cosine profile
    # Compact model
    mp.dm9.compact.NdmPad = ceil_even(1+2*mp.F3.Rin*mp.F3.compact.res)
    NbeamCompact = mp.dm9.compact.NdmPad
    mp.dm9.NdmPad = NbeamCompact
    mp.dm9.compact.Nbox = mp.dm9.compact.NdmPad # the modes take up the full array.
    # Normalized coordinates: Compact model
    if mp.centering == 'pixel':
        xc = np.arange(-mp.dm9.compact.NdmPad/2, mp.dm9.compact.NdmPad/2)/mp.F3.compact.res
    elif mp.centering == 'interpixel':
        xc = np.arange(-(mp.dm9.compact.NdmPad-1)/2, (mp.dm9.compact.NdmPad+1)/2)/mp.F3.compact.res
        
    [Xc, Yc] = np.meshgrid(xc, xc)
    Rc, THETAc = cart2pol(Xc, Yc)
    
    # Hyper-gaussian rolloff at edge of occulter
    hg_expon = 44  # Found empirically
    apRad = mp.F3.Rin/(mp.F3.Rin+0.1)  # Found empirically
    OD = 1
    mask = Rc <= mp.F3.Rin
    windowFull = mask*np.exp(-(Rc/mp.F3.Rin/(apRad*OD))**hg_expon)
    
    drSep = drCos/2
    min_azimSize = mp.min_azimSize_dm9  # [microns]
    pow_arr = np.arange(2, 62, 2)*6

    numdivCos = 2
    countCos = 0
    start_rad = int(np.floor(mp.dm9.actres/2))+1
    for ri in range(start_rad, Nrad+1):
        for _iter in range(numdivCos+1):
            countCos += 1
    numdivSin = 3
    countSin = 0
    start_rad = int(np.floor(mp.dm9.actres/2))+1
    for ri in range(start_rad, Nrad+1):
        for _iter in range(numdivSin+1):
            countSin += 1
    mp.dm9.NactTotal = Nrad + countCos + countSin
    mp.dm9.compact.inf_datacube = np.zeros((mp.dm9.compact.NdmPad,
                                    mp.dm9.compact.NdmPad, mp.dm9.NactTotal))
    
    # Compute the ring influence functions
    for ri in range(1, Nrad+1):
        modeTemp = windowFull * (1 + (-1)**((ri+1) % 2) *
                                 np.cos(2*np.pi*(Rc*mp.dm9.actres-0.5)))/2
        rMin = drSep*(ri - 1)
        rMax = drSep*(ri + 1)
        if ri == 1:  # Fill in the center
            modeTemp[Rc < drSep] = 1
        else:
            modeTemp[Rc < rMin] = 0
        modeTemp[Rc > rMax] = 0
        mp.dm9.compact.inf_datacube[:, :, ri-1] = modeTemp
        
    # for ri in range(1, Nrad+1):
    #     infFunc = mp.dm9.compact.inf_datacube[:, :, ri-1]
    #     plt.imshow(infFunc); plt.colorbar(); plt.pause(0.05)
    
    beamRad = NbeamCompact/2
    if mp.centering == 'pixel':
        x = np.arange(-beamRad, beamRad)
    elif mp.centering == 'interpixel':
        x = np.arange((NbeamCompact-1)/2, (NbeamCompact+1)/2)
    X, Y = np.meshgrid(x, x)
    RHO, THETA = cart2pol(X, Y)
    THETA2 = THETA+np.pi/3*2
    THETA3 = THETA+np.pi/3*4
    THETA4 = THETA+np.pi/3
    THETA5 = THETA+np.pi/3*3
    THETA6 = np.fliplr(THETA)
    apRad = mp.F3.Rin/(mp.F3.Rin+0.1)  # Found empirically
    OD = 1
    mask = Rc <= mp.F3.Rin
    
    # Cosine basis
    numdiv = 2
    count = 0  # index counter
    for ri in np.arange(start_rad, Nrad+1):
        modeTemp = windowFull * (1 + (-1)**((ri+1) % 2) *
                                 np.cos(2*np.pi*(Rc*mp.dm9.actres-0.5)))/2
        rMin = drSep*(ri - 1)
        rMax = drSep*(ri + 1)
        if ri == 1:  # Fill in the center
            modeTemp[Rc < drSep] = 1
        else:
            modeTemp[Rc < rMin] = 0
        
        modeTemp[Rc > rMax] = 0
        for II in range(numdiv+1):
            # Choose power for number of lobes
            powmin = 2 * np.pi * rMin / min_azimSize * 18
            aux = pow_arr - powmin
            aux[aux < 0] = np.Inf
            ind_mi = np.argmin(aux)
            power = pow_arr[ind_mi]
            #
            cosFull = np.cos(THETA*power) + 1
            numdiv = int(power/6)
            dth = 2*np.pi/power
            th_arr = np.linspace(np.pi/2, np.pi/2+np.pi/3, numdiv+1)
            th_rev_arr = np.linspace(np.pi/2+np.pi/3, np.pi/2, numdiv+1)
    
            th = th_arr[II]
            th_rev = th_rev_arr[II]
            ind = np.logical_and(THETA < (th+dth/2), THETA > (th-dth/2))
            ind_rev = np.logical_and(THETA4 < (th_rev+dth/2),
                                     THETA4 > (th_rev-dth/2))
            ind2 = np.logical_and(THETA2 < (th+dth/2),
                                  THETA2 > (th-dth/2))
            ind2_rev = np.logical_and(THETA5 < (th_rev+dth/2),
                                      THETA5 > (th_rev-dth/2))
            ind3 = np.logical_and(THETA3 < (th+dth/2), THETA3 > (th-dth/2))
            ind3_rev = np.logical_and(THETA6 < (th+dth/2-np.pi/2-np.pi/3/2),
                                      THETA6 > (th-dth/2-np.pi/2-np.pi/3/2))
            indTot = np.logical_or(ind, ind2)
            indTot = np.logical_or(indTot, ind3)
            indTot = np.logical_or(indTot, ind_rev)
            indTot = np.logical_or(indTot, ind2_rev)
            indTot = np.logical_or(indTot, ind3_rev)
            cosII = cosFull * indTot * modeTemp / 2
            mp.dm9.compact.inf_datacube[:, :, Nrad+count] = cosII
            count += 1

    # Sin basis
    numdiv = 3
    for ri in np.arange(start_rad, Nrad+1):
        modeTemp = windowFull * (1 + (-1)**((ri+1) % 2) *
                                 np.cos(2*np.pi*(Rc*mp.dm9.actres-0.5)))/2
        rMin = drSep*(ri - 1)
        rMax = drSep*(ri + 1)
        if ri == 1:  # Fill in the center
            modeTemp[Rc < drSep] = 1
        else:
            modeTemp[Rc < rMin] = 0
        modeTemp[Rc > rMax] = 0
        for II in range(numdiv+1):
            # Choose power for number of lobes
            powmin = 2*np.pi*rMin/min_azimSize*18
            aux = pow_arr - powmin
            aux[aux < 0] = np.Inf
            ind_mi = np.argmin(aux)
            power = pow_arr[ind_mi]
            # disp(['Number of lobes',num2str(pow)])
            cosFull = -np.cos(THETA*power)+1
            
            dth = 2*np.pi/power
            th_arr = np.linspace(np.pi/2, np.pi/2+np.pi/3+np.pi/6, numdiv+1)

            th = th_arr[II] + np.pi/12
            if th < np.pi:
                ind = np.logical_and(THETA < (th+dth/2), THETA > (th-dth/2))
            else:
                ind = np.fliplr(np.logical_and(THETA < (np.pi-th+dth/2),
                                               THETA > (np.pi-th-dth/2)))
            ind2 = np.logical_and(THETA2 < (th+dth/2),
                                  THETA2 > (th-dth/2))
            ind3 = np.logical_and(THETA3 < (th+dth/2),
                                  THETA3 > (th-dth/2))
            indTotsin = np.logical_or(ind, ind2)
            indTotsin = np.logical_or(indTotsin, ind3)
            cosII = cosFull * indTotsin * modeTemp / 2
            mp.dm9.compact.inf_datacube[:, :, Nrad+count] = cosII
            count += 1
            
    #         numdiv = int(power/6)
    #         dth = 2*np.pi/power
    #         th_arr = np.linspace(np.pi/2-np.pi/2/power, np.pi/2+np.pi/3-np.pi/2/power, numdiv+1)
    #         th_rev_arr = np.linspace(np.pi/2+np.pi/3+np.pi/2/power, np.pi/2+np.pi/2/power, numdiv+1)
    
    #         th = th_arr[II]
    #         th_rev = th_rev_arr[II]
    #         ind = np.logical_and((THETA)<(th+dth/2), (THETA)>(th-dth/2))
    #         ind_rev = np.logical_and((THETA4)<(th_rev+dth/2),
    #                                  (THETA4)>(th_rev-dth/2))
    #         ind2 = np.logical_and((THETA2)<(th+dth/2), (THETA2)>(th-dth/2))
    #         ind2_rev = np.logical_and((THETA5)<(th_rev+dth/2),
    #                                   (THETA5)>(th_rev-dth/2))
    #         ind3 = np.logical_and((THETA3)<(th+dth/2), (THETA3)>(th-dth/2))
    #         ind3_rev = np.logical_and((THETA6)<(th+dth/2-np.pi/2-np.pi/3/2),
    #                                   (THETA6)>(th-dth/2-np.pi/2-np.pi/3/2))
    #         indTot0 = np.logical_or(ind, ind2)
    #         indTot0 = np.logical_or(indTot0, ind3);
    #         indTot_rev = np.logical_or(ind_rev, ind2_rev)
    #         indTot_rev = np.logical_or(indTot_rev, ind3_rev)
    #     # %     cosII = zeros(N);
    #         sinII = sinFull*indTot0*modeTemp + sinFull_rev*indTot_rev*modeTemp
    # # %         figure(102);imagesc(sinII);axis image; set(gca,'YDir', 'normal')
    # # %         pause(0.1)
    #         mp.dm9.compact.inf_datacube[:, :, Nrad+count] = sinII
    #         count += 1
    
    mp.dm9.inf_datacube = mp.dm9.compact.inf_datacube
    
    mp.dm9.NactTotal = mp.dm9.inf_datacube.shape[2]
    mp.dm9.VtoH = mp.dm9.VtoHavg*np.ones(mp.dm9.NactTotal)
    
    # for ri in range(mp.dm9.NactTotal):
    #     infFunc = mp.dm9.compact.inf_datacube[:, :, ri]
    #     plt.imshow(infFunc); plt.colorbar(); plt.pause(0.01)
        
    # infSum = np.sum(mp.dm9.inf_datacube[:, :, Nrad+countCos:-1], 2)  # sin
    # infSum = np.sum(mp.dm9.inf_datacube[:, :, Nrad:Nrad+countCos], 2)  # cos
    # infSum = np.sum(mp.dm9.inf_datacube[:, :, 0:Nrad], 2)  # rings
    # infSum = np.sum(mp.dm9.inf_datacube[:, :, :], 2)  # all
    # plt.figure(); plt.imshow(infSum); plt.colorbar(); plt.pause(1)
    
    # Lower-left pixel coordinates are all (1,1) since the Zernikes take up the full array.
    mp.dm9.xy_box_lowerLeft = np.zeros((2, mp.dm9.NactTotal))
    mp.dm9.compact.xy_box_lowerLeft = np.zeros((2, mp.dm9.NactTotal))
    mp.dm9.compact.Nbox = NbeamCompact
    mp.dm9.Nbox = NbeamCompact
    
    # Coordinates for the full FPM array [meters]
    if mp.centering == 'pixel':
        mp.dm9.compact.x_pupPad = np.arange(-mp.dm9.compact.NdmPad/2,
                                            (mp.dm9.compact.NdmPad/2)) * \
                                    mp.dm9.compact.dxi
    elif mp.centering == 'interpixel':
        mp.dm9.compact.x_pupPad = np.arange(-(mp.dm9.compact.NdmPad-1)/2,
                                            (mp.dm9.compact.NdmPad+1)/2) * \
                                    mp.dm9.compact.dxi
    mp.dm9.compact.y_pupPad = mp.dm9.compact.x_pupPad
    
    # Initial DM9 voltages
    if not hasattr(mp.dm9, 'V'):
        mp.dm9.V = np.zeros(mp.dm9.NactTotal)
        mp.dm9.V[0:Nrad] = mp.dm9.V0coef * np.ones(Nrad)
    else:
        mp.dm9.V = mp.DM9V0
    mp.dm9.Vmin = np.min(mp.t_diel_nm_vec)  # minimum thickness of FPM dielectric layer (nm)
    mp.dm9.Vmax = np.max(mp.t_diel_nm_vec)  # maximum thickness (from one actuator, not of the facesheet) of FPM dielectric layer (nm)

    # OPTIONS FOR DEFINING DM8 (FPM Metal)
    mp.dm8.VtoHavg = 1e-9  # gain of DM8 (meters/Volt)
    mp.dm8.Vmin = np.min(mp.t_metal_nm_vec)  # minimum thickness of FPM metal layer (nm)
    mp.dm8.Vmax = np.max(mp.t_metal_nm_vec)  # maximum thickness (from one actuator, not of the facesheet) of FPM metal layer (nm)

    # DM8 Option 2: Set basis as a single nickel disk.
    mp.dm8.NactTotal = 1
    mp.dm8.act_ele = 1
    print('%d actuators in DM8.' % mp.dm8.NactTotal)
    mp.dm8.VtoH = mp.dm8.VtoHavg * np.ones(mp.dm8.NactTotal)  # Gains: volts to meters in surface height;
    mp.dm8.xy_box_lowerLeft = np.array([0, 0]).reshape((2, 1))
    mp.dm8.compact = mp.dm8
    if not hasattr(mp.dm8, 'V'):  # Initial DM8 voltages
        mp.dm8.V = mp.dm8.V0coef * np.ones(mp.dm8.NactTotal)
    else:
        mp.dm8.V = mp.DM8V0
        
    # Don't define extra actuators and time:
    if not mp.F3.Rin == mp.F3.RinA:
        raise ValueError('Change mp.F3.Rin and mp.F3.RinA to be equal to avoid wasting time.')
        
    # Copy over some common values from DM9:
    mp.dm8.dxi = mp.dm9.dxi  # Width of a pixel at the FPM in full model (meters)
    mp.dm8.NdmPad = mp.dm9.NdmPad
    mp.dm8.Nbox = mp.dm8.NdmPad
    mp.dm8.compact.dxi = mp.dm9.compact.dxi  # Width of a pixel at the FPM in compact model (meters)
    mp.dm8.compact.NdmPad = mp.dm9.compact.NdmPad
    mp.dm8.compact.Nbox = mp.dm8.compact.NdmPad

    # Make or read in DM8 disk for the full model
    FPMgenInputs = {}
    FPMgenInputs['pixresFPM'] = mp.F3.full.res  # pixels per lambda_c/D
    FPMgenInputs['rhoInner'] = mp.F3.Rin  # radius of inner FPM amplitude spot (in lambda_c/D)
    FPMgenInputs['rhoOuter'] = np.Inf  # radius of outer opaque FPM ring (in lambda_c/D)
    FPMgenInputs['FPMampFac'] = 0  # amplitude transmission of inner FPM spot
    FPMgenInputs['centering'] = mp.centering
    diskFull = np.round(pad_crop(1-falco.mask.gen_annular_FPM(FPMgenInputs),
                                 mp.dm8.NdmPad))
    mp.dm8.inf_datacube = np.zeros((diskFull.shape[0], diskFull.shape[1], 1))
    mp.dm8.inf_datacube[:, :, 0] = diskFull
    # Make or read in DM8 disk for the compact model
    FPMgenInputs['pixresFPM'] = mp.F3.compact.res  # pixels per lambda_c/D
    diskCompact = np.round(pad_crop(1-falco.mask.gen_annular_FPM(FPMgenInputs),
                                    mp.dm8.compact.NdmPad))
    mp.dm8.compact.inf_datacube = np.zeros((diskCompact.shape[0],
                                            diskCompact.shape[1], 1))
    mp.dm8.compact.inf_datacube[:, :, 0] = diskCompact
    pass


def setup_fpm(mp):

    mp.dm9.Nact = ceil_even(2*mp.F3.Rin*mp.dm9.actres)  # number of actuators across DM9 (if not in a hex grid)

    if mp.dm9.inf0name.lower() in '3x3':
        mp.dm9.inf0 = 1/4 * np.array([[1, 2, 1], [2, 4, 2], [1, 2, 1]])  # influence function
        mp.dm9.dx_inf0_act = 1/2  # number of inter-actuator widths per pixel 
        # FPM resolution (pixels per lambda0/D) in the compact and full models.
        mp.F3.compact.res = mp.dm9.actres / mp.dm9.dx_inf0_act
        mp.F3.full.res = mp.dm9.actres / mp.dm9.dx_inf0_act

    elif mp.dm9.inf0name.lower() in 'lanczos3':
        N = 91
        xs = np.arange(-(N-1)/2, (N+1)/2)/N*10*0.906  #(-(N-1)/2:(N-1)/2)/N *10*0.906

        a = 3
        Lx0 = a*np.sin(np.pi*xs)*np.sin(np.pi*xs/a)/(np.pi*xs)**2
        Lx0[xs == 0] = 1
        Lx = Lx0
        Lx[xs >= a] = 0
        Lx[xs <= -a] = 0
        Lxy = Lx.T @ Lx  # The 2-D Lanczos kernel
        Nhalf = np.ceil(N/2)
        Nrad = 30
        LxyCrop = Lxy[Nhalf-Nrad-1:Nhalf+Nrad, Nhalf-Nrad-1:Nhalf+Nrad]

        mp.dm9.inf0 = LxyCrop  # influence function
        mp.dm9.dx_inf0_act = 1/10  # number of inter-actuator widths per pixel 

    elif mp.dm9.inf0name.lower() in 'xinetics':
        mp.dm9.inf0 = 1*fits.getdata('influence_dm5v2.fits')
        mp.dm9.dx_inf0_act = 1/10  # number of inter-actuator widths per pixel 
    
    
    # DM8 and DM9 (Optimizable FPM) Setup

    # Centering of DM surfaces on array
    mp.dm8.centering = mp.centering
    mp.dm9.centering = mp.centering

    mp.dm9.compact = mp.dm9

    if hasattr(mp, 'flagDM9inf3x3'):
        mp.dm9.xcent_dm = mp.dm9.Nact/2 - 1/2
        mp.dm9.ycent_dm = mp.dm9.Nact/2 - 1/2
        if mp.centering in 'interpixel':
           raise ValueError('The 3x3 influence function for DM9 requires a pixel-centered coordinate system.')
    else:
        mp.dm9.xcent_dm = mp.dm9.Nact/2 - 1/2
        mp.dm9.ycent_dm = mp.dm9.Nact/2 - 1/2
    
    mp.dm9.dm_spacing = 1/mp.dm9.actres*(mp.fl*mp.lambda0/mp.P2.D)  # meters, pitch of DM actuators
    mp.dm9.compact = mp.dm9

    mp.dm9.compact.dm_spacing = mp.dm9.dm_spacing  # meters, pitch of DM actuators

    mp.dm9.dx_inf0 = (mp.dm9.dx_inf0_act)*mp.dm9.dm_spacing  # meters, sampling of the influence function 

    mp.dm9.compact.dx_inf0 = (mp.dm9.compact.dx_inf0_act)*mp.dm9.compact.dm_spacing  # meters, sampling of the influence function 

    mp.dm9.dxi = (mp.fl*mp.lambda0/mp.P2.D)/mp.F3.full.res  # width of a pixel at the FPM in the full model (meters)
    mp.dm9.compact.dxi = (mp.fl*mp.lambda0/mp.P2.D)/mp.F3.compact.res  # width of a pixel at the FPM in the compact model (meters)

    if mp.dm9.inf0.shape[0] == 3:
        fpm_inf_cube_3x3(mp.dm9)
        fpm_inf_cube_3x3(mp.dm9.compact)
    else:
        fpm_inf_cube(mp.dm9)
        fpm_inf_cube(mp.dm9.compact)
        
    # Zero out DM9 actuators too close to the outer edge (within mp.dm9.FPMbuffer lambda0/D of edge)
    r_cent_lam0D = mp.dm9.r_cent_act*mp.dm9.dm_spacing/(mp.dm9.dxi)/mp.F3.full.res

    ##
    mp.F3.RinA_inds = np.array([], dtype=int)
    mp.F3.RinAB_inds = np.array([], dtype=int)
    for ii in range(mp.dm9.NactTotal):
        # Zero out FPM actuators beyond the allowed radius (mp.F3.Rin)
        if r_cent_lam0D[ii] > mp.F3.Rin-mp.dm9.FPMbuffer:
           mp.dm9.inf_datacube[:, :, ii] = np.zeros_like(mp.dm9.inf_datacube[:, :, ii])
           mp.dm9.compact.inf_datacube[:, :, ii] = np.zeros_like(mp.dm9.compact.inf_datacube[:, :, ii])

        # Get the indices for the actuators within radius mp.F3.RinA
        if r_cent_lam0D[ii] <= mp.F3.RinA-mp.dm9.FPMbuffer:
            mp.F3.RinA_inds = np.append(mp.F3.RinA_inds, ii)
        else:  # Get the indices for the actuators between radii mp.F3.RinA and mp.F3.Rin
            mp.F3.RinAB_inds = np.append(mp.F3.RinAB_inds, ii)

    print('%d actuators in DM9.' % mp.dm9.NactTotal)

    mp.dm9.ABfac = 1  # Gain factor between inner and outer FPM regions
    mp.dm9.VtoHavg = 1e-9  # gain of DM9 (meters/Volt)
    mp.dm9.VtoH = mp.dm9.VtoHavg * np.ones(mp.dm9.NactTotal)  # Gains: volts to meters in surface height;
    mp.dm9.VtoH[mp.F3.RinAB_inds] = mp.dm9.ABfac * mp.dm9.VtoH[mp.F3.RinAB_inds]

    if not hasattr(mp.dm9, 'V'):  # Initial DM9 voltages
        mp.dm9.V = np.zeros(mp.dm9.NactTotal)
        mp.dm9.V[mp.F3.RinA_inds] = mp.dm9.V0coef
    else:
        mp.dm9.V = mp.DM9V0

    mp.dm9.Vmin = np.min(mp.t_diel_nm_vec)  # minimum thickness of FPM dielectric layer (nm)
    mp.dm9.Vmax = np.max(mp.t_diel_nm_vec)  # maximum thickness (from one actuator, not of the facesheet) of FPM dielectric layer (nm)

    # -OPTIONS FOR DEFINING DM8 (FPM Metal)
    mp.dm8.VtoHavg = 1e-9  # gain of DM8 (meters/Volt)
    mp.dm8.Vmin = np.min(mp.t_metal_nm_vec)  # minimum thickness of FPM metal layer (nm)
    mp.dm8.Vmax = np.max(mp.t_metal_nm_vec)  # maximum thickness (from one actuator, not of the facesheet) of FPM metal layer (nm)

    # DM8 Option 2: Set basis as a single nickel disk.
    mp.dm8.NactTotal = 1
    mp.dm8.act_ele = 1
    print('%d actuators in DM8.' % mp.dm8.NactTotal)
    mp.dm8.VtoH = mp.dm8.VtoHavg * np.ones(mp.dm8.NactTotal)  # Gains: volts to meters in surface height;
    mp.dm8.xy_box_lowerLeft = np.array([0, 0]).reshape((2, 1))
    mp.dm8.compact = mp.dm8
    if not hasattr(mp.dm8, 'V'):  # Initial DM8 voltages
        mp.dm8.V = mp.dm8.V0coef * np.ones(mp.dm8.NactTotal)
    else:
        mp.dm8.V = mp.DM8V0
        
    # Don't define extra actuators and time:
    if not mp.F3.Rin == mp.F3.RinA:
        raise ValueError('Change mp.F3.Rin and mp.F3.RinA to be equal to avoid wasting time.')
        
    # Copy over some common values from DM9:
    mp.dm8.dxi = mp.dm9.dxi  # Width of a pixel at the FPM in full model (meters)
    mp.dm8.NdmPad = mp.dm9.NdmPad
    mp.dm8.Nbox = mp.dm8.NdmPad
    mp.dm8.compact.dxi = mp.dm9.compact.dxi  # Width of a pixel at the FPM in compact model (meters)
    mp.dm8.compact.NdmPad = mp.dm9.compact.NdmPad
    mp.dm8.compact.Nbox = mp.dm8.compact.NdmPad

    # Make or read in DM8 disk for the full model
    FPMgenInputs = {}
    FPMgenInputs['pixresFPM'] = mp.F3.full.res  # pixels per lambda_c/D
    FPMgenInputs['rhoInner'] = mp.F3.Rin  # radius of inner FPM amplitude spot (in lambda_c/D)
    FPMgenInputs['rhoOuter'] = np.Inf  # radius of outer opaque FPM ring (in lambda_c/D)
    FPMgenInputs['FPMampFac'] = 0  # amplitude transmission of inner FPM spot
    FPMgenInputs['centering'] = mp.centering
    diskFull = np.round(pad_crop(1-falco.mask.gen_annular_FPM(FPMgenInputs),
                                 mp.dm8.NdmPad))
    mp.dm8.inf_datacube = np.zeros((diskFull.shape[0], diskFull.shape[1], 1))
    mp.dm8.inf_datacube[:, :, 0] = diskFull
    # Make or read in DM8 disk for the compact model
    FPMgenInputs['pixresFPM'] = mp.F3.compact.res  # pixels per lambda_c/D
    diskCompact = np.round(pad_crop(1-falco.mask.gen_annular_FPM(FPMgenInputs),
                                    mp.dm8.compact.NdmPad))
    mp.dm8.compact.inf_datacube = np.zeros((diskCompact.shape[0],
                                            diskCompact.shape[1], 1))
    mp.dm8.compact.inf_datacube[:, :, 0] = diskCompact
    
    # Zero out parts of DM9 actuators that go outside the nickel disk.
    # Also apply the grayscale edge.
    DM8windowFull = diskFull
    DM8windowCompact = diskCompact
    for iact in range(mp.dm9.NactTotal):
        if np.sum(np.abs(mp.dm9.inf_datacube[:, :, iact])) >= 1e-8 and np.abs(mp.dm9.VtoH[iact]) >= 1e-13:
            y_box_ind = np.arange(mp.dm9.xy_box_lowerLeft[0, iact], mp.dm9.xy_box_lowerLeft[0, iact]+mp.dm9.Nbox, dtype=int)  # x-indices in pupil arrays for the box
            x_box_ind = np.arange(mp.dm9.xy_box_lowerLeft[1, iact], mp.dm9.xy_box_lowerLeft[1, iact]+mp.dm9.Nbox, dtype=int)  # y-indices in pupil arrays for the box
            mp.dm9.inf_datacube[:, :, iact] = DM8windowFull[np.ix_(x_box_ind, y_box_ind)] * mp.dm9.inf_datacube[:, :, iact]

            y_box_ind = np.arange(mp.dm9.compact.xy_box_lowerLeft[0, iact], mp.dm9.compact.xy_box_lowerLeft[0, iact]+mp.dm9.compact.Nbox, dtype=int)  # x-indices in pupil arrays for the box
            x_box_ind = np.arange(mp.dm9.compact.xy_box_lowerLeft[1, iact], mp.dm9.compact.xy_box_lowerLeft[1, iact]+mp.dm9.compact.Nbox, dtype=int)  # y-indices in pupil arrays for the box
            mp.dm9.compact.inf_datacube[:, :, iact] = DM8windowCompact[np.ix_(x_box_ind, y_box_ind)] * mp.dm9.compact.inf_datacube[:, :, iact]

    pass


def gen_fpm_poke_cube(dm, mp, dx_dm, **kwargs):
    """
    Compute the datacube of each influence function.
    
    Influence functions are cropped down or padded up
    to the best size for angular spectrum propagation.

    Parameters
    ----------
    dm : ModelParameters
        Structure containing parameter values for the DM
    mp: falco.config.ModelParameters
        Structure of model parameters
    dx_dm : float
        Pixel width [meters] at the DM plane

    Other Parameters
    ----------------
    NOCUBE : bool
       Switch that tells function not to compute the datacube of influence
       functions.

    Returns
    -------
    None
        modifies structure "dm" by reference

    """
    if type(mp) is not falco.config.ModelParameters:
        raise TypeError('Input "mp" must be of type ModelParameters')
    real_positive_scalar(dx_dm, 'dx_dm', TypeError)
          
    if "NOCUBE" in kwargs and kwargs["NOCUBE"]:
        flagGenCube = False
    else:
        flagGenCube = True

    # Define this flag if it doesn't exist in the older code for square actuator arrays only
    if not hasattr(dm, 'flag_hex_array'):
        dm.flag_hex_array = False

    # Set the order of operations
    XYZ = True
    if(hasattr(dm, 'flagZYX')):
        if(dm.flagZYX):
            XYZ = False

    # Compute sampling of the pupil. Assume that it is square.
    dm.dx_dm = dx_dm
    dm.dx = dx_dm

    # Default to being centered on a pixel if not specified
    if not(hasattr(dm, 'centering')):
        dm.centering = 'pixel'
        
    # Compute coordinates of original influence function
    Ninf0 = dm.inf0.shape[0]  # Number of points across inf func at native res
    x_inf0 = np.linspace(-(Ninf0-1)/2., (Ninf0-1)/2., Ninf0)*dm.dx_inf0
    # True for even- or odd-sized influence function maps as long as they are
    # centered on the array.
    [Xinf0, Yinf0] = np.meshgrid(x_inf0, x_inf0)
    
    # Number of points across the DM surface at native inf func resolution
    Ndm0 = falco.util.ceil_even(Ninf0 + (dm.Nact - 1)*(dm.dm_spacing/dm.dx_inf0))
    # Number of points across the (un-rotated) DM surface at new, desired res.
    dm.NdmMin = falco.util.ceil_even(Ndm0*(dm.dx_inf0/dm.dx))+2.
    # Number of points across the array to fully contain the DM surface at new
    # desired resolution and z-rotation angle.
    dm.Ndm = int(ceil_even((abs(np.array([np.sqrt(2.)*np.cos(radians(45.-dm.zrot)),
            np.sqrt(2.)*np.sin(radians(45.-dm.zrot))])).max())*Ndm0*(dm.dx_inf0/dm.dx))+2)
    
    # Compute list of initial actuator center coordinates (in actutor widths).
    if(dm.flag_hex_array):  # Hexagonal, hex-packed grid
        raise ValueError('flag_hex_array option not implemented yet.')
#     Nrings = dm.Nrings;
#     x_vec = [];
#     y_vec = [];
#     % row number (rowNum) is 1 for the center row and 2 is above it, etc.
#     % Nacross is the total number of segments across that row
#     for rowNum = 1:Nrings
#         Nacross = 2*Nrings - rowNum  # Number of actuators across at that row (for hex tiling in a hex shape)
#         yval = sqrt(3)/2*(rowNum-1);
#         bx = Nrings - (rowNum+1)/2  # x offset from origin
# 
#         xs = (0:Nacross-1).' - bx  # x values are 1 apart
#         ys = yval*ones(Nacross,1)  # same y-value for the entire row
# 
#         if(rowNum==1)
#             x_vec = [x_vec;xs];
#             y_vec = [y_vec;ys]; 
#         else
#             x_vec = [x_vec;xs;xs];
#             y_vec = [y_vec;ys;-ys]  # rows +/-n have +/- y coordinates
#         end
#     end
    else:  # Square grid [actuator widths]
        [dm.Xact, dm.Yact] = np.meshgrid(np.arange(dm.Nact) - 
                                         dm.xc, np.arange(dm.Nact)-dm.yc)
#        # Use order='F' to compare the final datacube to Matlab's output.
#        #  Otherwise, use C ordering for Python FALCO.
#        x_vec = dm.Xact.reshape(dm.Nact*dm.Nact,order='F')
#        y_vec = dm.Yact.reshape(dm.Nact*dm.Nact,order='F')
        x_vec = dm.Xact.reshape(dm.Nact*dm.Nact)
        y_vec = dm.Yact.reshape(dm.Nact*dm.Nact)
    
    dm.NactTotal = x_vec.shape[0]  # Total number of actuators in the 2-D array
    dm.xy_cent_act = np.zeros((2, dm.NactTotal))  # Initialize

    # Compute the rotation matrix to apply to the influence function and 
    #  actuator center locations
    tlt = np.zeros(3)
    tlt[0] = radians(dm.xtilt)
    tlt[1] = radians(dm.ytilt)
    tlt[2] = radians(-dm.zrot)

    sa = sin(tlt[0])
    ca = cos(tlt[0])
    sb = sin(tlt[1])
    cb = cos(tlt[1])
    sg = sin(tlt[2])
    cg = cos(tlt[2])

    if XYZ:
        Mrot = np.array([[cb * cg, sa * sb * cg - ca * sg, ca * sb * cg + sa * sg, 0.0],
                 [cb * sg, sa * sb * sg + ca * cg, ca * sb * sg - sa * cg, 0.0],
                [-sb,      sa * cb,                ca * cb,                0.0],
                     [0.0,                    0.0,                    0.0, 1.0]])
    else:
        Mrot = np.array([               [cb * cg,               -cb * sg,       sb, 0.0],
                [ca * sg + sa * sb * cg, ca * cg - sa * sb * sg, -sa * cb, 0.0],
                [sa * sg - ca * sb * cg, sa * cg + ca * sb * sg,  ca * cb, 0.0],
                                   [0.0,                    0.0,      0.0, 1.0]])
    
    # Compute the actuator center coordinates in units of actuator spacings
    for iact in range(dm.NactTotal):
        xyzVals = np.array([x_vec[iact], y_vec[iact], 0., 1.])
        xyzValsRot = Mrot @ xyzVals
        dm.xy_cent_act[0, iact] = xyzValsRot[0].copy()
        dm.xy_cent_act[1, iact] = xyzValsRot[1].copy()

    N0 = dm.inf0.shape[0]
    Npad = falco.util.ceil_odd(np.sqrt(2.)*N0)
    inf0pad = np.zeros((Npad, Npad))
    inf0pad[int(np.ceil(Npad/2.)-np.floor(N0/2.)-1):int(np.ceil(Npad/2.)+np.floor(N0/2.)),
            int(np.ceil(Npad/2.)-np.floor(N0/2.)-1):int(np.ceil(Npad/2.)+np.floor(N0/2.))] = dm.inf0

    ydim = inf0pad.shape[0]
    xdim = inf0pad.shape[1]
    
    xd2 = np.fix(xdim / 2.) + 1
    yd2 = np.fix(ydim / 2.) + 1
    cx = np.arange(xdim) + 1. - xd2
    cy = np.arange(ydim) + 1. - yd2
    [Xs0, Ys0] = np.meshgrid(cx, cy)

    xsNewVec = np.zeros(xdim*xdim)
    ysNewVec = np.zeros(ydim*ydim)
    Xs0Vec = Xs0.reshape(xdim*xdim)
    Ys0Vec = Ys0.reshape(ydim*ydim)
    
    for ii in range(Xs0.size):
        xyzVals = np.array([Xs0Vec[ii], Ys0Vec[ii], 0., 1.])
        xyzValsRot = Mrot @ xyzVals
        xsNewVec[ii] = xyzValsRot[0]
        ysNewVec[ii] = xyzValsRot[1]
    
    # Calculate the interpolated DM grid at the new resolution
    # (set extrapolated values to 0.0)
    dm.infMaster = griddata((xsNewVec, ysNewVec), inf0pad.reshape(Npad*Npad),
                            (Xs0, Ys0), method='cubic', fill_value=0.)

    # Crop down the influence function until it has no zero padding left
    infSum = np.sum(dm.infMaster)
    infDiff = 0.
    counter = 0
    while(abs(infDiff) <= 1e-7):
        counter = counter + 2
        infDiff = infSum - np.sum(abs(dm.infMaster[int(counter/2):int(-counter/2),
                                                   int(counter/2):int(-counter/2)]))

    # Subtract an extra 2 to negate the extra step that overshoots.
    counter = counter - 2
    Ninf0pad = dm.infMaster.shape[0]-counter
    if counter == 0:
        infMaster2 = dm.infMaster.copy()
    else:
        # The cropped-down influence function
        infMaster2 = dm.infMaster[int(counter/2):int(-counter/2),
                                  int(counter/2):int(-counter/2)].copy()
        dm.infMaster = infMaster2
    
    Npad = Ninf0pad

    # True for even- or odd-sized influence function maps as long as they are
    # centered on the array.
    x_inf0 = np.linspace(-(Npad-1)/2, (Npad-1)/2., Npad)*dm.dx_inf0
    [Xinf0, Yinf0] = np.meshgrid(x_inf0, x_inf0)

    # Translate and resample the master influence function to be at each 
    # actuator's location in the pixel grid 

    # Compute the size of the postage stamps.
    # Number of points across the influence function array at the DM plane's
    # resolution. Want as even
    Nbox = falco.util.ceil_even(Ninf0pad*dm.dx_inf0/dx_dm)
    dm.Nbox = Nbox
    # Also compute their padded sizes for the angular spectrum (AS) propagation
    # between P2 and DM1 or between DM1 and DM2
    # Minimum number of points across for accurate angular spectrum propagation
    Nmin = falco.util.ceil_even(np.max(mp.sbp_centers)*np.max(np.abs(np.array(
        [mp.d_P2_dm1, mp.d_dm1_dm2, (mp.d_P2_dm1+mp.d_dm1_dm2)])))/dx_dm**2)
    # Use a larger array if the max sampling criterion for angular spectrum
    # propagation is violated
    dm.NboxAS = np.max(np.array([Nbox, Nmin]))

    # Pad the pupil to at least the size of the DM(s) surface(s) to allow all
    # actuators to be located outside the pupil.
    # (Same for both DMs)

    # Find actuator farthest from center:
    dm.r_cent_act = np.sqrt(dm.xy_cent_act[0, :]**2 + dm.xy_cent_act[1, :]**2)
    dm.rmax = np.max(np.abs(dm.r_cent_act))
    NpixPerAct = dm.dm_spacing/dx_dm
    if(dm.flag_hex_array):
        # padded 2 actuators past the last actuator center to avoid trying to
        # index outside the array
        dm.NdmPad = falco.util.ceil_even((2.*(dm.rmax+2))*NpixPerAct + 1)
    else:
        # DM surface array padded by the width of the padded influence function
        # to prevent indexing outside the array. 
        # The 1/2 term is because the farthest actuator center is still half an
        # actuator away from the nominal array edge. 
        dm.NdmPad = falco.util.ceil_even((dm.NboxAS + 2.0*(1 + (np.max(
        np.abs(dm.xy_cent_act.reshape(2*dm.NactTotal)))+0.5)*NpixPerAct)))

    # Compute coordinates (in meters) of the full DM array
    if(dm.centering == 'pixel'):
        # meters, coords for the full DM arrays. Origin is centered on a pixel
        dm.x_pupPad = np.linspace(-dm.NdmPad/2., (dm.NdmPad/2. - 1),
                                  dm.NdmPad)*dx_dm
    else:
        # meters, coords for the full DM arrays. Origin is interpixel centered
        dm.x_pupPad = np.linspace(-(dm.NdmPad-1)/2., (dm.NdmPad-1)/2.,
                                  dm.NdmPad)*dx_dm

    dm.y_pupPad = dm.x_pupPad

    # Make NboxPad-sized postage stamps for each actuator's influence function
    if(flagGenCube):
        if not dm.flag_hex_array:
            print("  Influence function padded from %d to %d points for A.S. propagation." % (Nbox,dm.NboxAS))

        print('Computing datacube of DM influence functions... ', end='')

        # Find the locations of the postage stamps arrays in the larger pupilPad array
        dm.xy_cent_act_inPix = dm.xy_cent_act*(dm.dm_spacing/dx_dm)  # Convert units to pupil-plane pixels
        dm.xy_cent_act_inPix = dm.xy_cent_act_inPix + 0.5  # For the half-pixel offset if pixel centered. 
        dm.xy_cent_act_box = np.round(dm.xy_cent_act_inPix)  # Center locations of the postage stamps (in between pixels), in actuator widths
        dm.xy_cent_act_box_inM = dm.xy_cent_act_box*dx_dm  # now in meters
        dm.xy_box_lowerLeft = dm.xy_cent_act_box + (dm.NdmPad-Nbox)/2 - 0  # index of pixel in lower left of the postage stamp within the whole pupilPad array. +0 for Python, +1 for Matlab

        # Starting coordinates (in actuator widths) for updated master influence function.
        # This is interpixel centered, so do not translate!
        dm.x_box0 = np.linspace(-(Nbox-1)/2., (Nbox-1)/2., Nbox)*dx_dm
        [dm.Xbox0, dm.Ybox0] = np.meshgrid(dm.x_box0, dm.x_box0)  # meters, interpixel-centered coordinates for the master influence function

        # (Allow for later) Limit the actuators used to those within 1 actuator width of the pupil
        r_cent_act_box_inM = np.sqrt(dm.xy_cent_act_box_inM[0, :]**2 + dm.xy_cent_act_box_inM[1, :]**2)
        # Compute and store all the influence functions:
        dm.inf_datacube = np.zeros((Nbox, Nbox, dm.NactTotal))  # initialize array of influence function "postage stamps"
        dm.act_ele = np.arange(dm.NactTotal)  # Initialize as including all actuators

        inf_datacube = np.zeros((dm.NactTotal, Nbox, Nbox))

        interp_spline = RectBivariateSpline(x_inf0, x_inf0, dm.infMaster)  # RectBivariateSpline is faster in 2-D than interp2d
        # Refer to https://scipython.com/book/chapter-8-scipy/examples/two-dimensional-interpolation-with-scipyinterpolaterectbivariatespline/

        for iact in range(dm.NactTotal):
           xbox = dm.x_box0 - (dm.xy_cent_act_inPix[0, iact]-dm.xy_cent_act_box[0, iact])*dx_dm # X = X0 -(x_true_center-x_box_center)
           ybox = dm.x_box0 - (dm.xy_cent_act_inPix[1, iact]-dm.xy_cent_act_box[1, iact])*dx_dm # Y = Y0 -(y_true_center-y_box_center)
           dm.inf_datacube[:, :, iact] = interp_spline(ybox, xbox)
           inf_datacube[iact, :, :] = interp_spline(ybox, xbox)

        print('done.')

    else:
        dm.act_ele = np.arange(dm.NactTotal)
   

def fpm_inf_cube_3x3(dm):
    print('Computing datacube of FPM influence functions... ')
    
    # Compute sampling of the pupil. Assume that it is square.
    dm.dx_dm = dm.dxi
    
    # Default to being centered on a pixel (FFT-style)
    if not hasattr(dm, 'centering'):
        dm.centering = 'pixel'
    
    # Compute coordinates of original influence function
    Ninf0 = dm.inf0.shape[0]
    x_inf0 = np.arange(-(Ninf0-1)/2, (Ninf0+1)/2) * dm.dx_inf0  # True for even- or odd-sized influence function maps as long as they are centered on the array.
    [Xinf0, Yinf0] = np.meshgrid(x_inf0, x_inf0)
    
    # Compute list of initial actuator center coordinates (in actuator widths).
    # Square grid
    [dm.Xact, dm.Yact] = np.meshgrid(np.arange(0, dm.Nact)-dm.xcent_dm, np.arange(0, dm.Nact)-dm.ycent_dm)  # in actuator widths
    x_vec = dm.Xact.flatten()
    y_vec = dm.Yact.flatten()
    
    dm.NactTotal = x_vec.size  # Total number of actuators in the 2-D array
    
    dm.infMaster = dm.inf0
    Nbox = dm.inf0.shape[0]
    dm.Nbox = Nbox
    print('FPM influence function size =\t%dx%d ' % (Nbox, Nbox))
    
    dm.xy_cent_act = np.vstack((x_vec, y_vec))  # in actuator widths
    
    # Pad the pupil to at least the size of the DM(s) surface(s) to allow all
    # actuators to be located outside the pupil.
    # (Same for both DMs)
    
    # Find actuator farthest from center:
    dm.r_cent_act = np.sqrt(dm.xy_cent_act[0, :]**2 + dm.xy_cent_act[1, :]**2)
    dm.rmax = np.max(np.abs(dm.r_cent_act))
    dm.absxymax = np.max(np.abs(dm.xy_cent_act))
    NpixPerActWidth = dm.dm_spacing / dm.dx_dm
    
    dm.NdmPad = ceil_even((dm.Nact+2)*NpixPerActWidth)  # prevent indexing outside the array
    
    # Compute coordinates (in meters) of the full DM array
    if dm.centering in 'pixel':
        dm.x_pupPad = np.arange(-dm.NdmPad/2, dm.NdmPad/2)*dm.dx_dm  # meters, coords for the full DM arrays. Origin is centered on a pixel
    else:
        dm.x_pupPad = np.arange(-(dm.NdmPad-1)/2, (dm.NdmPad+1)/2)*dm.dx_dm  # meters, coords for the full DM arrays. Origin is centered between pixels for an even-sized array

    dm.y_pupPad = dm.x_pupPad
    
    # DM: (use NboxPad-sized postage stamps,
    
    # Find the locations of the postage stamps arrays in the larger pupilPad array
    dm.xy_cent_act_inPix = dm.xy_cent_act*(dm.dm_spacing/dm.dx_dm)  # Convert units to pupil-file pixels
    if not dm.centering in 'pixel':
        raise ValueError('Not adapted for non-pixel centering.')
    
    dm.xy_cent_act_box = np.round(dm.xy_cent_act_inPix)  # Center locations of the postage stamps (in between pixels), in actuator widths
    dm.xy_cent_act_box_inM = dm.xy_cent_act_box*dm.dx_dm  # now in meters
    if Nbox % 2 == 0:
        dm.xy_box_lowerLeft = dm.xy_cent_act_box + (dm.NdmPad-Nbox)/2  # indices of pixel in lower left of the postage stamp within the whole pupilPad array
    else:
        dm.xy_box_lowerLeft = dm.xy_cent_act_box + (dm.NdmPad)/2 - np.floor(Nbox/2)  # indices of pixel in lower left of the postage stamp within the whole pupilPad array
    
    # Starting coordinates (in actuator widths) for updated influence function. This is
    # interpixel centered, so do not translate!
    dm.x_box0 = np.arange(-(Nbox-1)/2, (Nbox+1)/2) * dm.dx_dm
    [dm.Xbox0, dm.Ybox0] = np.meshgrid(dm.x_box0, dm.x_box0)  # meters, interpixel-centered coordinates for the master influence function
    
    # Limit the actuators used to those within 1 actuator width of the pupil
    r_cent_act_box_inM = np.sqrt(dm.xy_cent_act_box_inM[0, :]**2 + dm.xy_cent_act_box_inM[1, :]**2)
    # Compute and store all the influence functions:
    dm.inf_datacube = np.zeros((Nbox, Nbox, dm.NactTotal))  # initialize array of influence function "postage stamps"
    dm.act_ele = np.array([])  # Indices of nonzero-ed actuators
    for iact in range(dm.NactTotal):
       dm.act_ele = np.append(dm.act_ele, iact)  # Add actuator index to the keeper list
       dm.inf_datacube[:, :, iact] = dm.inf0
    
    print('done.')
    
    
def gen_fpm(mp):  #RENAME

    # Number of points across the FPM in the full model
    mp.F3.full.Nxi = mp.dm9.NdmPad
    mp.F3.full.Neta = mp.F3.full.Nxi
    # Number of points across the FPM in the compact model
    mp.F3.compact.Nxi = mp.dm9.compact.NdmPad
    mp.F3.compact.Neta = mp.F3.compact.Nxi

    # Starting DM8 and DM9 surfaces
    if mp.flagPlot:
        DM8surf = gen_fpm_surf_from_cube(mp.dm8, 'full')
        plt.figure(); plt.imshow(DM8surf); plt.colorbar(); plt.title('FPM Metal Thickness'); plt.pause(1)

        DM9surf = gen_fpm_surf_from_cube(mp.dm9, 'full')
        plt.figure(); plt.imshow(DM9surf); plt.colorbar(); plt.title('FPM Dielectric Thickness'); plt.pause(1)
        
    pass

def fpm_inf_cube(dm):
    pass